!<CPS_v1 type="PROCESS"/>

!<-- This documentation header was last revised by CI Burch on 2000-09-08. />

!<license>
!-------------------------------------------------------------------------------
! Copyright (c) 2007 ConocoPhillips Company
!
! Permission is hereby granted, free of charge, to any person obtaining a copy
! of this software and associated documentation files (the "Software"), to deal
! in the Software without restriction, including without limitation the rights
! to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
! copies of the Software, and to permit persons to whom the Software is
! furnished to do so, subject to the following conditions:
!
! The above copyright notice and this permission notice shall be included in all
! copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
! AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
! SOFTWARE.
!-------------------------------------------------------------------------------
!</license>

!<brief_doc>
!-------------------------------------------------------------------------------
!                         C P S   P R O C E S S                  
!
! Name       : CADZOW    (F-X Cadzow coherence enhancement)
! Category   : filters
! Written    : 2017-11-11   by: Michael
!			based on FXDECON (Chiu and Stoekley)
! Revised    : 2018-05-05
! 	     : 
! Maturity   : beta
! Purpose    : 2D coherence enhancement by SVD of frequency slices
! Portability: No known limitations.
! Parallel   : No.
!
!-------------------------------------------------------------------------------
!</brief_doc>

!<descript_doc>
!-------------------------------------------------------------------------------
!                          GENERAL DESCRIPTION                   
!
! CADZOW coherence enhancement is done separately for each time window for
! each input trace.  Time windows overlap by 50%.
! 
! Each trace subset of NUM_TR traces is Fourier transformed over time to form 
! the F-X window.  Then a trajectory matrix is made from the frequency slice.
! SVD is performed and a reduced number of components used to preserve the
! coherent events.  The anti-diagonals of the approximate matrix are averaged
! and then put back into the frequency slice. This algorithm is
! repeated for each frequency independently, and the processed volume is
! inverse Fourier transformed to give the output trace time window. 
! It can preserve amplitudes better than traditional f-x decon, but will
! require more CPU time.
!-------------------------------------------------------------------------------
!</descript_doc>

!<advice_doc>
!-------------------------------------------------------------------------------
!                            ADVICE FOR USERS                      
!
! CADZOW is a 2D process only. It can be applied post-stack or to shots.
!
!  Noise Attenuation and Parameter Choice
! For shallow hi-res data, try a shorter WIN_LEN such as 0.25 seconds.
! For deep target, WIN_LEN of 0.6 seconds may be better. For most oil/gas
! seismic, 0.4 to 0.5 seconds is suitable.
! WIN_LEN should be at least 40 samples (0.16 seconds at 4 ms sampling).
! More traces in window increases runtime. For SVD, this is a function of
! N cubed. N is actually NUM_TR+2*OVERLAP (only NUM_TR+OVERLAP for first
! windows of a gather.) Increasing SVD_RANK will also increase the runtime
! a little: rank 5 will take 20% longer than rank 2.
! For D dips in a window, SVD_RANK must be at least D.
! rank 2 will remove most noise, but is only suitable for "tramlines".
! rank 5 will preserve most signal but remove less noise. If attempting
! SVD_RANK=2, then OVERLAP=6 is suggested to minimise amplitude errors.
! Maximum reduced rank will be (NUM_TR+OVERLAP-1)/2
! For pre-stack data, rank 5 will generally be safe, and rank 4 for stacked
! data. For noisy data with simple geology, rank 3 may be expedient. If NMO is
! applied, you may get away with lower rank. If CADZOW creates noise on the
! first and last trace of a gather, that may be a symptom that the rank is
! less than number of dips present.
! It is not effective against swell noise, which may appear as locally coherent.
! Works best against random noise such as wind on geophones.
! Note the theory assumes data is made of linear events. So a normal moveout
! correction, even approximate, may help preserve shallow reflections.
! Dead traces should not be deleted.  CADZOW will interpolate dead traces,
! if they are at least 20 traces apart. This interpolation is better with
! lower rank. Otherwise in common-offset gathers, missing shots can be
! interpolated with SPTI, which is rather crude, then CADZOW will improve
! the interpolated traces.
!-------------------------------------------------------------------------------
!</advice_doc>
!<trace_in_doc>
!-------------------------------------------------------------------------------
!                      TRACE INPUT REQUIREMENTS                
!
!
! Process is a multiple-trace process.
! Traces can be input one at a time, or in gathers (single trace support used).
!
!-------------------------------------------------------------------------------
!</trace_in_doc>

!<trace_out_doc>
!-------------------------------------------------------------------------------
!                     TRACE OUTPUT CHARACTERISTICS             
!
!
! This process alters input traces.
! This process outputs the same traces as it receives (possibly altered).
!
! This process outputs one trace at a time.
!
!-------------------------------------------------------------------------------
!</trace_out_doc>

!<global_doc>
!-------------------------------------------------------------------------------
!       PROJECT DATA, JOB DATA, AND GLOBAL PARAMETERS USED OR CHANGED      
!
! Name      Description                             Action taken
! ----      -----------                             ------------
! NUMTR     max number of traces input/output       used but not changed
! NWIH      number of words in trace header         used but not changed
! NDPT      number of sample values in trace        used but not changed
! TSTRT     starting time on trace                  used but not changed
! DT        trace sample interval                   used but not changed
!
!-------------------------------------------------------------------------------
!</global_doc>

!<header_word_doc>
!-------------------------------------------------------------------------------
!                   TRACE HEADER WORDS USED OR CHANGED           
! 
! Hwd#    Description                Action taken
! ----    -----------                ------------
!
! 2       Head mute index            Used
! 25      LAV                        Reset
! 64      Tail mute index            Used
!         HDR_LINE                   Used
!       
!-------------------------------------------------------------------------------
!</header_word_doc>

!<history_doc>
!-------------------------------------------------------------------------------
!                             REVISION HISTORY                    
! 
!     Date        Author      Description
!     ----        ------      -----------
! 12  2018-05-01			Addback 50% of first trace
! 11. 2017-11-11	      	Modify FXDECON to create this process.
! 10. 2017-??				Force WIN_LEN to have even samples
! 9.  2016-11-26			Fixed cadzow_sol() running off ends of cdata1
!							array for last group of gather.
! 8.  2010-08-02  Stoeckley   Modify to allow multiple trace input.  This change
!                              means that the original oneset execution subroutine
!                              cadzow was renamed cadzow_single_trace, a new twosets
!                              subroutine was created which calls cadzow_single_trace,
!                              and the restriction of the global NUMTR being 1 was
!                              removed.
! 7.  2010-01-12  Stoeckley   Add pc_error messages if not single trace input;
!                             add return statement in main execution routine
!                             if not single trace input.
! 6.  2009-11-18  Bill Menger Added test for divide by zero that was causing NaN's
!                             in the output on rare occasions.
! 5.  2006-01-03  S.Chiu      Fix problem when autocorrelation is zero
! 4.  2004-02-05  S.Chiu      Fix problem when the number of traces in a CMP 
!                             gather is less than NUM_TR.
! 3.  2002-05-07  S.Chiu      Add option of KEEP_LO_FREQ_DATA.
! 2.  2001-04-02  S.Chiu      Change wrapped_up to skip_wrapup.
! 1.  2000-11-29  S.Chiu      Initial version
!
!-------------------------------------------------------------------------------
!</history_doc>

!<portability_doc>
!-------------------------------------------------------------------------------
!                        PORTABILITY LIMITATIONS               
!
! Tested on linux only.
!
!-------------------------------------------------------------------------------
!</portability_doc>

!<compile_doc>
!-------------------------------------------------------------------------------
!                      SPECIAL COMPILING REQUIREMENTS             
!
! Tested only with GNU gfortran and Intel ifort.
!
!-------------------------------------------------------------------------------
!</compile_doc>

!<calling_doc>
!-------------------------------------------------------------------------------
!                    SPECIFIC CALLING CHARACTERISTICS       
! Control
! Parameter     Value
! Name          Reported   Description
! ----          --------   -----------
! NTAPES           0       number of magnetic tapes needed.
! NEED_REQUEST   true      whether this process ever needs to request traces.
! NEED_LABEL     true      whether this process needs a label.     
! TWOSETS        true      whether this process needs two trace/header arrays.
! NSCRATCH         0       amount of temporary memory needed.       
! NSTORE           0       amount of permanent memory needed.      
! IFTD           false     whether this process frees tape drives. 
! NDISK            0       disk space needed (megabytes) if large. 
! SETUP_ONLY     false     whether this process is setup-only.    
!
! Upon input, NTR must have one of these values:
!  NTR >= 1              means to process the input traces.
!  NTR == NO_MORE_TRACES means there are no more input traces.
!  NTR == NEED_TRACES    means someone else needs more traces.
!
! Upon output, NTR will have one of these values:
!  NTR >= 1              if this process is outputting traces.
!  NTR == NO_MORE_TRACES if there are no more traces to output.
!  NTR == FATAL_ERROR    if this process has a fatal error.
!  NTR == NEED_TRACES    if this process needs more traces.
!
!-------------------------------------------------------------------------------
!</calling_doc>

!<int_calling_doc>
!-------------------------------------------------------------------------------
!                    ALTERNATE INTERNAL CALLING METHODS          
!
!  None provided.
!
!-------------------------------------------------------------------------------
!</int_calling_doc>

!<algorithm_doc>
!-------------------------------------------------------------------------------
!                   ALGORITHM DESCRIPTION FOR DEVELOPERS         
! Cadzow, J. A. 1988: IEEE Trans. on Acoustics, Speech and Signal Processing
! vol 36, p 49-62
! Only one iteration is done, so the method is equivalent to Singular Spectrum
! Analysis, which is well known in various fields other than seismology.
! For the first trace of an ensemble, 50% of original trace is added back if
! rank < 5, because filter does not help first trace.
!-------------------------------------------------------------------------------
!</algorithm_doc>

!<programming_doc>
!-------------------------------------------------------------------------------
!                            PROGRAMMING NOTES                 
!
! Making matrix square is optimum but restricts some parameters to odd/even.
!
!-------------------------------------------------------------------------------
!</programming_doc>
!-------------------------------------------------------------------------------
!
!-------------------------------------------------------------------------------
!<--  XML code for the GUI goes in this section. />
!
!-------------------------------------------------------------------------------
!
!<gui_def>
!<NS CADZOW Process/NC=80>
!
!        F-X noise attenuation for regularly sampled traces.
!
! HDR_LINE= `IIIIII       OVERLAP= `IIIIII      NUM_TR=~~ `IIIIIII
!
! WIN_LEN=~~`FFFFFFFF    FREQ_BEG= `FFFFFFFF    FREQ_END= `FFFFFFF
!
! SVD_RANK= `IIIII       KEEP_LO_FREQ_DATA =`CCCCC
!
!</gui_def>
!
!-------------------------------------------------------------------------------
!<--  Parameter help information goes in this section. />
!
!<HelpSection>
!
!<Help KEYWORD="HDR_LINE">
!<Tip> Header word defining input lines or other trace panels. </Tip>
! Default = 3
! Allowed = 1 - NWIH
! HDR_LINE should be a different whole number for each distinct input line or 
! other trace panel. Suggest 3,7,8,9 or 10
!</Help>
!
!<Help KEYWORD="OVERLAP">
!<Tip> Number of traces overlap as window moves. </Tip>
! Default = 4
! Allowed = 2,4,6 EVEN
! Overlap is needed since attenuation is less at edge? 
! More overlap increases runtime.
!</Help>
!
!<Help KEYWORD="NUM_TR">
!<Tip> Number of filtered traces for the window. </Tip>
! Default = 13 
! Allowed = int 9 - 27 ODD
! Moving window advances by NUM_TR
! Runtime increases with wider window
!  
!</Help>
!
!<Help KEYWORD="WIN_LEN">
!<Tip> Length, in seconds, of the local time window. </Tip>
! Default = 0.5
! Allowed = real > 0.1 and even number of samples
! Each trace group is broken into overlapping time windows of length WIN_LEN 
! seconds for F-X filtering. Actual FFT window will be at least 20% longer.
!</Help>
!
!<Help KEYWORD="FREQ_BEG">
!<Tip> Minimum frequency for the F-X filter. </Tip>
! Default = 2.0
! Allowed = real >= 0.0
! See also comments for KEEP_LO_FREQ_DATA.
!</Help>
!
!<Help KEYWORD="FREQ_END">
!<Tip> Maximum frequency for the F-X filter. </Tip>
! Default = Nyquist*0.75
! Allowed = Nyquist >= real > FREQ_BEG
!</Help>
!
!<Help KEYWORD="KEEP_LO_FREQ_DATA">
!<Tip> Keep original data from frquency 0 to FREQ_BEG-FREQ_INC </Tip>
! Default = NO
! Allowed = YES/NO  
! If KEEP_LO_FREQ_DATA = YES, then no FX noise attenuation applies
! to data from frquency 0 to FREQ_BEG-FREQ_INC. This helps to keep 
! the low-frequency fault plane reflections. 
! FREQ_INC = Nyquist frequency /(length of the FFT/2+1) 
! If KEEP_LO_FREQ_DATA = NO, then FX noise attenuation applies
! taper to data from frequency 0 to FREQ_BEG-FREQ_INC.
!</Help>
!
!<Help KEYWORD="SVD_RANK">
!<Tip> Rank of matrix after reduction by SVD. </Tip>
! Default = 4
! Allowed = integer > 1
! This is the number of principal components kept by the filter. Increasing
! rank will pass more dips in the data, but reduce the noise rejection.
! Maximum rank allowed is (NUM_TR+OVERLAP-1)/2
!</Help>
!
!
!</HelpSection>

!-------------------------------------------------------------------------------
!
!!--------------------------- start of module ------------------------------!!
!!--------------------------- start of module ------------------------------!!
!!--------------------------- start of module ------------------------------!!


      module cadzow_module

      use singletracesupport_module
      use pc_module
      use named_constants_module
      use getlun_module
      use string_module
      use lav_module
      Use sort_module
      use fft_module
      use mutehw_module
      use gausselim_module

      implicit none
      private
      public :: cadzow_create
      public :: cadzow_initialize
      public :: cadzow_update
      public :: cadzow_delete
!<execute_only>
      public :: cadzow            ! main execution (trace processing) routine.
      public :: cadzow_wrapup
!</execute_only>


      character(len=100),public,save :: cadzow_IDENT = &
       '$Id: cazdow.f90,v 1.2 2018/05/01 12:36:22 S.Chiu prod sps $'

!!---------------------- parameter structure -------------------------------!!
!!---------------------- parameter structure -------------------------------!!
!!---------------------- parameter structure -------------------------------!!
                                    

    type,public :: cadzow_struct
 
    private

    logical               :: skip_wrapup         ! wrapup flag.
!   logical               :: gathered           ! gathered flag.
 
    integer               :: hdr_line           ! Process parameters
    integer               :: num_pts            ! Process parameters
    integer               :: num_tr             ! Process parameters
    real                  :: freq_beg           ! Process parameters
    real                  :: freq_end           ! Process parameters
    real                  :: win_len            ! Process parameters
    logical               :: keep_lo_freq_data  ! Process parameters
    integer		  		:: svd_rank		! Process parameters
   
    integer               :: ndpt               ! globals
    integer               :: nwih               ! globals
    real                  :: dt                 ! globals
    real                  :: tstrt              ! globals
    
    logical               :: ifirst             ! dependent variables
    logical               :: ilast              ! dependent variables
    logical               :: eodata             ! dependent variables
    integer               :: done               ! dependent variables
    integer               :: ieout              ! dependent variables
    integer               :: ifreq1             ! dependent variables
    integer               :: ifreq2             ! dependent variables
    integer               :: iout               ! dependent variables
    integer               :: ioinpt             ! dependent variables
    integer               :: isout              ! dependent variables
    integer               :: itrseq             ! dependent variables
    integer               :: hdval              ! dependent variables
    integer               :: hdval_old          ! dependent variables
    integer               :: neq                ! NOT USED
    integer               :: nfftp2             ! dependent variables
    integer               :: nhftp1             ! dependent variables
    integer               :: nout               ! dependent variables
    integer               :: ntr_panel          ! dependent variables
    integer               :: ntpts              ! dependent variables
    integer               :: nwin               ! dependent variables
    integer               :: nxstor             ! dependent variables
    integer               :: ntr_gather         ! dependent variables
    integer               :: nread              ! dependent variables
    integer               :: print_lun          ! dependent variables.
    real                  :: pwhite             ! NOT USED
   
    double precision, pointer      :: hd_save(:)    ! dependent variables
    double precision, pointer      :: hdstor(:,:)   ! dependent variables
    real,             pointer      :: tr_save(:)    ! dependent variables
    real,             pointer      :: trstor(:,:)   ! dependent variables
    integer,          pointer      :: indx_tr(:)    ! dependent variables
    integer,          pointer      :: iseq(:)       ! dependent variables
    real,             pointer      :: outtr(:,:)    ! dependent variables
    real,             pointer      :: outwt(:,:)    ! dependent variables
    real,             pointer      :: winwt(:)      ! dependent variables

    type(fft_struct),pointer :: fftrc_obj           ! dependent variables.
    type(fft_struct),pointer :: fftcr_obj           ! dependent variables.

    type(singletracesupport_struct),pointer :: sts

!   integer                     :: itr_in           ! for single trace I/O support.
!   integer                     :: itr_out          ! for single trace I/O support.
!   integer                     :: ntr_inout        ! for single trace I/O support.
!   double precision, pointer   :: hd_single(:,:)   ! for single trace I/O support.
!   real, pointer               :: tr_single(:,:)   ! for single trace I/O support.
!   logical                     :: finished         ! for single trace I/O support.

    end type cadzow_struct

!!--------------------------------- data -----------------------------------!!
!!--------------------------------- data -----------------------------------!!
!!--------------------------------- data -----------------------------------!!


      type(cadzow_struct),pointer,save :: object      ! needed for traps.

      contains


!!----------------------------- create -------------------------------------!!
!!----------------------------- create -------------------------------------!!
!!----------------------------- create -------------------------------------!!


      subroutine cadzow_create (obj)
      implicit none
      type(cadzow_struct),pointer :: obj       ! arguments

      allocate (obj)

!    Nullify ALL POINTERS in your parameter structure 

      nullify  (obj%hd_save)
      nullify  (obj%hdstor)
      nullify  (obj%tr_save)
      nullify  (obj%trstor)
      nullify  (obj%indx_tr)
      nullify  (obj%iseq)
      nullify  (obj%outtr) 
      nullify  (obj%outwt) 
      nullify  (obj%winwt)
      nullify  (obj%fftrc_obj)
      nullify  (obj%fftcr_obj)
      nullify  (obj%sts)
!     nullify  (obj%hd_single)  ! for single trace I/O support.
!     nullify  (obj%tr_single)  ! for single trace I/O support.

      call cadzow_initialize (obj)
      return
      end subroutine cadzow_create


!!------------------------------- delete -----------------------------------!!
!!------------------------------- delete -----------------------------------!!
!!------------------------------- delete -----------------------------------!!


      subroutine cadzow_delete (obj)
      implicit none
      type(cadzow_struct),pointer :: obj       ! arguments

!<execute_only>
      call cadzow_wrapup (obj)
!</execute_only>

!! Make sure ALL POINTERS in parameter structure are deallocated

      if (associated(obj%hd_save)) deallocate (obj%hd_save)
      if (associated(obj%hdstor))  deallocate (obj%hdstor)
      if (associated(obj%tr_save)) deallocate (obj%tr_save)
      if (associated(obj%trstor))  deallocate (obj%trstor)
      if (associated(obj%indx_tr)) deallocate (obj%indx_tr)
      if (associated(obj%iseq))    deallocate (obj%iseq)
      if (associated(obj%outtr))   deallocate (obj%outtr)
      if (associated(obj%outwt))   deallocate (obj%outwt)
      if (associated(obj%winwt))   deallocate (obj%winwt)

      if (associated(obj%fftrc_obj))  call fft_delete (obj%fftrc_obj)
      if (associated(obj%fftcr_obj))  call fft_delete (obj%fftcr_obj)

!     if (associated(obj%tr_single)) deallocate (obj%tr_single)  ! for single trace I/O support.
!     if (associated(obj%hd_single)) deallocate (obj%hd_single)  ! for single trace I/O support.

      deallocate(obj)
      return
      end subroutine cadzow_delete


!!----------------------------- initialize ---------------------------------!!
!!----------------------------- initialize ---------------------------------!!
!!----------------------------- initialize ---------------------------------!!


      subroutine cadzow_initialize (obj)
      implicit none
      type(cadzow_struct),intent(inout) :: obj         ! arguments

      real                               :: rnyquist    ! local

!     Initialize ALL NON-POINTER VARIABLES in parameter structure

      call pc_get_global ('dt'    , obj%dt)    ! trace sample interval (s).
      rnyquist = (1.0/(2*obj%dt))
 
      obj%hdr_line  = 3
      obj%freq_beg  = 2.0
      obj%freq_end  = AINT(rnyquist*0.75)
      obj%num_pts   = 4
      obj%num_tr    = 13
      obj%win_len   = 0.5
      obj%svd_rank  = 4
      obj%keep_lo_freq_data = .false.

      obj%print_lun = pc_get_lun()
      call cadzow_update (obj)
      return
      end subroutine cadzow_initialize


!!------------------------- start of update --------------------------------!!
!!------------------------- start of update --------------------------------!!
!!------------------------- start of update --------------------------------!!


      subroutine cadzow_update (obj)
      implicit none
      type(cadzow_struct),intent(inout),target :: obj               ! arguments
   

!! Declare all required local variables as follows:

!     integer             :: numtr,ntapes,nscratch,nstore,ndisk      ! local
      integer             :: ntapes,nscratch,nstore,ndisk            ! local
      logical             :: iftd                                    ! local
      character(len=3)    :: need_label,need_request                 ! local
      character(len=3)    :: twosets                                 ! local    

      integer             :: nyquist, nfft                           ! local 
      integer             :: ier, len, i, m                          ! local
      real                :: tmid,  dist                             ! local
 

      object => obj               ! needed for traps.
      obj%skip_wrapup = .true.    ! needed for the wrapup routine.


!!------------------------- read parameters --------------------------------!!
!!------------------------- read parameters --------------------------------!!
!!------------------------- read parameters --------------------------------!!


!     numtr = inil
!     call pc_get_global ('numtr'   ,numtr)        ! maximum number of traces.
!     call pc_get_global ('gathered',obj%gathered) ! whether properly gathered.
      call pc_get_global ('nwih'  , obj%nwih)  ! number of header words.
      call pc_get_global ('ndpt'  , obj%ndpt)  ! number of trace samples.
      call pc_get_global ('tstrt' , obj%tstrt) ! time of 1st trace sample (s).
      call pc_get_global ('dt'    , obj%dt)    ! trace sample interval (s).

!------------Check that globals are set:

!     if (numtr == inil) call pc_error ("NUMTR global hasn't been set.")
      if (obj%nwih == inil) call pc_error ("NWIH global hasn't been set.")
      if (obj%ndpt == inil) call pc_error ("NDPT global hasn't been set.")
      if (obj%tstrt == fnil) call pc_error ("TSTRT global hasn't been set.")
      if (obj%dt   == fnil) call pc_error ("DT global hasn't been set.")
!     if (numtr > 1)       call pc_error ('CADZOW requires single trace input.')  ! no longer a restriction.

      call pc_get ('HDR_LINE',  obj%hdr_line)
      call pc_get ('FREQ_BEG',  obj%freq_beg)
      call pc_get ('FREQ_END',  obj%freq_end)
      call pc_get ('OVERLAP',   obj%num_pts)
      call pc_get ('NUM_TR',    obj%num_tr)
      call pc_get ('WIN_LEN',   obj%win_len)
      call pc_get ('SVD_RANK',  obj%svd_rank)
      call pc_get ('KEEP_LO_FREQ_DATA', obj%keep_lo_freq_data)
 
    !!------------------------- verify parameters --------------------------!!
    !!------------------------- verify parameters --------------------------!!
    !!------------------------- verify parameters --------------------------!!


     nyquist = nint(1.0/(2*obj%dt))
     if (obj%hdr_line < 1 .or. obj%hdr_line > obj%nwih) then   
       call pc_error (' CADZOW error: HDR_LINE must be between 1 and ', &
                            obj%nwih)
	obj%hdr_line=3
     end if 
     
     if (obj%freq_beg < 0.0 .or. obj%freq_beg > nyquist) then   
       call pc_error (' CADZOW error: FREQ_BEG must be between 0 and ', &
                        nyquist)
	obj%freq_beg=2.0
     end if

     if (obj%freq_end < obj%freq_beg ) then
       call pc_error (' CADZOW error: FREQ_END must be >= FREQ_BEG ')
     end if 

     if (obj%freq_end > nyquist ) then
       call pc_error (' CADZOW error: FREQ_END must be <= Nyquist of ', &
                        nyquist)
	obj%freq_end=AINT(nyquist*0.9)
     end if 

     if (obj%num_pts < 2 .or. obj%num_pts > 6 ) then
        call pc_error (' CADZOW error: OVERLAP must be between 2 and 6 ')
	obj%num_pts=4
     end if 
     if (mod(obj%num_pts,2) .ne. 0) then
        call pc_error (' CADZOW error: OVERLAP must be even number ')
	obj%num_pts=4
     end if 

     if (mod(obj%num_tr,2) == 0) then
        call pc_error (' CADZOW error: NUM_TR must be odd number ')
	obj%num_tr=13
     end if 
     if (obj%num_tr < 9 .or. obj%num_tr > 27 ) then
       call pc_error (' CADZOW error: NUM_TR must be between 9 and 27')
	obj%num_tr=13
     end if 

     m=(obj%num_tr+obj%num_pts-1)/2	! matrix size -1
     if(obj%svd_rank < 2) then
       call pc_error (' CADZOW error: SVD_RANK must be > 1')
	obj%svd_rank = 2
     end if
     if(obj%svd_rank > m) then
       call pc_error (' CADZOW error: SVD_RANK too large')
	obj%svd_rank = m
     end if

     if (obj%win_len <= 0.1 ) then
       call pc_error (' CADZOW error: WIN_LEN must be > 0.1')
	obj%win_len=0.16
     end if 
     if (obj%win_len > (obj%ndpt-1)*obj%dt ) then
       call pc_error (' CADZOW error: WIN_LEN must be <= trace length of ', &
                        (obj%ndpt-1)*obj%dt)
	obj%win_len = (obj%ndpt-1)*obj%dt
     end if
     obj%ntpts = nint(obj%win_len/obj%dt + 0.499) 
     if (mod(obj%ntpts,2) == 1) then   
        call pc_error (' CADZOW error: WIN_LEN must be even samples')
	obj%ntpts=obj%ntpts+1
	if (obj%ntpts > obj%ndpt ) obj%ntpts=obj%ntpts-2	! borderline case
	obj%win_len=obj%ntpts*obj%dt
     end if 

!!-------------------------- verify parameters -----------------------------!!
!!-------------------------- verify parameters -----------------------------!!
!!-------------------------- verify parameters -----------------------------!!



!!----------------------- write parameters ---------------------------------!!
!!----------------------- write parameters ---------------------------------!!
!!----------------------- write parameters ---------------------------------!!


!     call pc_put_global ('numtr'   , 1)          ! now numtr can be > 0 and is not changed.
!     call pc_put_global ('gathered', .false. ) 

      call pc_put ('HDR_LINE',  obj%hdr_line)
      call pc_put ('FREQ_BEG',  obj%freq_beg)
      call pc_put ('FREQ_END',  obj%freq_end)
      call pc_put ('OVERLAP',   obj%num_pts)
      call pc_put ('NUM_TR',    obj%num_tr)
      call pc_put ('WIN_LEN',   obj%win_len)
      call pc_put ('SVD_RANK',  obj%svd_rank)
      call pc_put ('KEEP_LO_FREQ_DATA', obj%keep_lo_freq_data)

!!----------------------- prepare for execution ----------------------------!!
!!----------------------- prepare for execution ----------------------------!!
!!----------------------- prepare for execution ----------------------------!!

!! Conditionally deallocate all arrays 
  
      if (associated(obj%hd_save)) deallocate (obj%hd_save)
      if (associated(obj%hdstor))  deallocate (obj%hdstor)
      if (associated(obj%tr_save)) deallocate (obj%tr_save)
      if (associated(obj%trstor))  deallocate (obj%trstor)
      if (associated(obj%indx_tr)) deallocate (obj%indx_tr)
      if (associated(obj%iseq))    deallocate (obj%iseq)
      if (associated(obj%outtr))   deallocate (obj%outtr)
      if (associated(obj%outwt))   deallocate (obj%outwt)
      if (associated(obj%winwt))   deallocate (obj%winwt)

      if (associated(obj%fftrc_obj))  call fft_delete (obj%fftrc_obj)
      if (associated(obj%fftcr_obj))  call fft_delete (obj%fftcr_obj)

!     if (associated(obj%tr_single)) deallocate (obj%tr_single)  ! for single trace I/O support.
!     if (associated(obj%hd_single)) deallocate (obj%hd_single)  ! for single trace I/O support.

      call singletracesupport_delete (obj%sts)

!     obj%itr_in    = 0         ! for single trace I/O support.
!     obj%itr_out   = 0         ! for single trace I/O support.
!     obj%ntr_inout = 0         ! for single trace I/O support.
!     obj%finished  = .false.   ! for single trace I/O support.

!   obj%nout     --  number of output traces.
!   obj%isout    --  starting index of output traces.
!   obj%ieout    --  starting index of output traces.
!   obj%iout     --  current index of output traces.
!   obj%nxstor   --  number of stored traces in memory buffer.
!   obj%trstor   --  stored traces in memory buffer.
!   obj%hdstor   --  stored headers in memory buffer.
!   obj%indx_tr  --  sorted of sequential index of the circular
!   obj%outtr    --  output interpolated traces in memory buffer.
!   obj%outwt    --  weights of traces in memory buffer.
!   obj%pwhite   --  prewhitening.

!.... Preset some initial values

      obj%eodata = .false.
      obj%ifirst = .true.
      obj%ilast  = .false.
      obj%done = 0 
      obj%ioinpt = 1 
      obj%ntr_panel = 0 
      obj%itrseq = 0  
      obj%nread = obj%num_tr + obj%num_pts

      obj%pwhite = 0.01
      obj%iout = 0 
 
      obj%ntpts = min(obj%ndpt,obj%ntpts) 
 
      obj%nxstor = obj%num_tr + obj%num_pts*2  
      obj%ntpts = nint(obj%win_len/obj%dt + 0.499) 
      if (obj%ntpts*1.5 > obj%ndpt) obj%ntpts = obj%ndpt 
 
!..... fft in mixed radix
 
      nfft = fft_nfctr(obj%ntpts*6/5)  
      obj%nfftp2 = nfft + 2  
      obj%nhftp1 = obj%nfftp2/2 

      nyquist = nint(1./(2*obj%dt)) 
      obj%ifreq1 = nint(obj%freq_beg*obj%nfftp2/(2*nyquist)) 
      obj%ifreq2 = nint(obj%freq_end*obj%nfftp2/(2*nyquist)) 
      obj%ifreq1 = max(obj%ifreq1,1) 
      obj%ifreq2 = min(obj%ifreq2,obj%nfftp2/2) 
 
      nstore = (obj%nwih+1)*obj%nxstor + (4*obj%ndpt+3)*obj%nxstor 
      nscratch = 1          
 
      need_label   = 'YES'
      need_request = 'YES'
      twosets      = 'YES'     ! for single trace I/O support.
      iftd         = .false.
      ndisk        = 0
      ntapes       = 0 
               
      call pc_put_control ('nstore',             nstore)
      call pc_put_control ('nscratch',         nscratch)
      call pc_put_control ('need_label',     need_label)
      call pc_put_control ('need_request', need_request)
      call pc_put_control ('twosets',           twosets)


!<execute_only>

      if (pc_do_not_process_traces()) return
      obj%skip_wrapup = .false.

!! Allocate your permanent memory:

      allocate  (obj%hd_save(obj%nwih),             stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW hd_save array')
 
      allocate  (obj%hdstor(obj%nwih, obj%nxstor),  stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW hdstor array')

      allocate  (obj%tr_save(obj%ndpt),             stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW tr_save array')

      allocate  (obj%trstor(obj%ndpt, obj%nxstor),  stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW trstor array')

      allocate  (obj%indx_tr(obj%nxstor),              stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW hd_save array')

      allocate  (obj%iseq(obj%nxstor),                  stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW iseq array')

      allocate  (obj%outtr(obj%ndpt, obj%nxstor ),     stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW outtr array')
 
      allocate  (obj%outwt(obj%ndpt, obj%nxstor),      stat=ier) 
      if (ier/=0) call pc_error ('Error creating CADZOW outwt array')

      allocate  (obj%winwt(obj%ntpts),  stat=ier)
      if (ier/=0) call pc_error ('Error creating CADZOW winwt array')

      call singletracesupport_create (obj%sts, obj%nwih, obj%ndpt)

!     allocate (obj%hd_single(obj%nwih,1), stat=ier)   ! for single trace I/O support.
!     allocate (obj%tr_single(obj%ndpt,1), stat=ier)   ! for single trace I/O support.

!.... initialize fft tables.
      ier =  fft_create (obj%fftrc_obj,  1, nfft, 'rtoc')
      if (ier/=0) call pc_error ('Error creating rtoc FFT object')
      ier =  fft_create (obj%fftcr_obj, -1, nfft, 'ctor')
      if (ier/=0) call pc_error ('Error creating ctor FFT object')

!.... initialize the iseq array to increment by 1

      obj%iseq = (/(i, i=1,obj%nxstor)/) 
 
!..... compute number of overlapped windows.
 
       len = obj%win_len/obj%dt + 1
       obj%nwin = nint(1.*obj%ndpt/len + 0.4999)  
       obj%nwin = 2*obj%nwin -1 
       if (mod(obj%ndpt,len) <= nint(0.5*len)) then
         obj%nwin = obj%nwin -1
       end if

!
!....  compute the weights of the a time window.
! 
      tmid = 0.5*(1. + obj%ntpts)  
      dist  = 1./(1.01*(obj%ntpts - tmid))
      do i = 1, obj%ntpts 
        obj%winwt(i) = 1. - abs((i - tmid))*dist 
        obj%winwt(i) =  obj%winwt(i)* obj%winwt(i)
      end do 
 
      if (pc_do_not_process_traces()) return   ! in case of allocation errors.

!</execute_only>


!!------------------------- finish update ----------------------------------!!
!!------------------------- finish update ----------------------------------!!
!!------------------------- finish update ----------------------------------!!


      return
      end subroutine cadzow_update


!!------------------------------- traps ------------------------------------!!
!!------------------------------- traps ------------------------------------!!
!!------------------------------- traps ------------------------------------!!


!!--------------------------- main execution -------------------------------!!
!!--------------------------- main execution -------------------------------!!
!!--------------------------- main execution -------------------------------!!



subroutine cadzow (obj, ntr, hdi, tri, hdo, tro)
    implicit none
    type(cadzow_struct), intent(inout) :: obj
    integer         , intent(inout) :: ntr
    double precision, intent(inout) :: hdi(:,:)
    real            , intent(inout) :: tri(:,:)
    double precision, intent(out)   :: hdo(:,:)
    real            , intent(out)   :: tro(:,:)

!    if (ntr == NO_MORE_TRACES) then
!        continue  ! do nothing here.
!    else if (ntr == NEED_TRACES) then
!        if (obj%finished) then    ! the last gather has already been output.
!            ntr = NO_MORE_TRACES
!            return
!        endif
!        obj%itr_out = 0   ! need to start refilling the output gather.
!    else
!        obj%ntr_inout = ntr
!        obj%itr_in = 1
!        obj%hd_single(1:obj%nwih,1) = hdi(1:obj%nwih,obj%itr_in)
!        obj%tr_single(1:obj%ndpt,1) = tri(1:obj%ndpt,obj%itr_in)
!        ntr = 1
!    endif

     if (singletracesupport_input(obj%sts, ntr, hdi, tri)) return

 22  call cadzow_single_trace (obj, ntr, hdi, tri)

     if (singletracesupport_output(obj%sts, ntr, hdi, tri, hdo, tro)) go to 22

!22  call cadzow_single_trace (obj, ntr, obj%hd_single, obj%tr_single)
!
!    if (ntr > 0) then
!        obj%itr_out = obj%itr_out + 1
!        hdo(1:obj%nwih,obj%itr_out) = obj%hd_single(1:obj%nwih,1)
!        tro(1:obj%ndpt,obj%itr_out) = obj%tr_single(1:obj%ndpt,1)
!        if (obj%itr_out == obj%ntr_inout) then
!            ntr = obj%ntr_inout    ! return a full output gather.
!            obj%itr_out = 0      ! will begin to refill the output gather.
!            return
!        endif
!        ntr = NEED_TRACES
!        goto 22
!    else if (ntr == FATAL_ERROR) then
!        return
!    else if (ntr == NO_MORE_TRACES) then  ! return what remains of the output gather, or 0.
!        if (obj%itr_out == 0) return   ! no more traces to output.
!        obj%finished = .true.
!        ntr = obj%itr_out    ! outputting whatever is in the last output gather ever.
!        return
!    else if (ntr == NEED_TRACES) then
!        if (obj%itr_in == obj%ntr_inout) return  ! go get more traces since the input gather is used up.
!        obj%itr_in = obj%itr_in + 1
!        obj%hd_single(1:obj%nwih,1) = hdi(1:obj%nwih,obj%itr_in)
!        obj%tr_single(1:obj%ndpt,1) = tri(1:obj%ndpt,obj%itr_in)
!        ntr = 1
!        goto 22
!    endif

end subroutine cadzow


!!----------------------- cadzow single trace -----------------------------!!
!!----------------------- cadzow single trace -----------------------------!!
!!----------------------- cadzow single trace -----------------------------!!


!! Upon input, NTR will have one of these values:
!!   NTR >= 1              means to process the input traces.
!!   NTR == NO_MORE_TRACES means there are no more imput traces.
!!   NTR == NEED_TRACES    means someone from below needs more traces.
!!   NTR == NEED_TRACES    might mean this is a trace-supplying process.
!!   NTR == NEED_TRACES    will not occur unless this process has a label.
!!
!! Upon output, NTR must have one of these values:
!!   NTR >= 1              if you are outputting traces.
!!   NTR == NO_MORE_TRACES if there are no more traces to output.
!!   NTR == FATAL_ERROR    if you have a fatal error.
!!   NTR == NEED_TRACES    if you need another trace before passing any out.
!!   NTR == NEED_TRACES    must not occur unless you specified that you
!!                           might need to request more traces.
!!
!<execute_only>


!! Need one set  of traces and headers, 

      subroutine cadzow_single_trace (obj, ntr, hd, tr)
      implicit none
      type(cadzow_struct),intent(inout) :: obj                ! arguments

      integer          ,intent(inout) :: ntr                   ! arguments
      double precision ,intent(inout) :: hd(:,:)               ! arguments
      real             ,intent(inout) :: tr(:,:)               ! arguments

      integer          ::  ipos, ierr, i, iadd                 ! local
        
      if  ( (ntr== NO_MORE_TRACES .and. obj%done==1 .and.       & 
             obj%iout>=obj%nout) .or. ntr == FATAL_ERROR )  then
         call cadzow_wrapup (obj)
         return
      end if
 
!.... ioinpt = 1 -- input phase

      if (obj%ioinpt == 1) then 
          if (ntr > 1) then
             call pc_error ('CADZOW requires single trace input.')
             ntr = FATAL_ERROR
             return
          end if

!....     first trace in the gather
          if (obj%itrseq == 0) obj%hdval_old = nint(hd(obj%hdr_line,ntr))
              
          if (ntr /= 0) then 
            obj%hdval = nint(hd(obj%hdr_line,ntr))  
          else 
            obj%hdval = -99999 
          endif 
 
!..... gather a panel of traces.
 
          if (ntr/=0 .and. obj%hdval_old==obj%hdval) then 
            obj%ntr_panel = obj%ntr_panel + 1 
            obj%itrseq = obj%itrseq + 1 
            ipos = mod(obj%itrseq,obj%nxstor) 
            if (ipos == 0) ipos = obj%nxstor 
            obj%iseq(ipos) = obj%itrseq 

  910       format('itrseq,ntr_panel,ipos,ntr,ioinpt,nxstor,hd7-8 ',6i7,2f9.1)
! first time, ipos=1
             obj%hdstor(1:obj%nwih, ipos) = hd(1:obj%nwih,ntr) 
             obj%trstor(1:obj%ndpt, ipos) = tr(1:obj%ndpt,ntr)

            ierr  = 0
            if (obj%ntr_panel == obj%nread) then

!..... Apply FX Cadzow Filter.
 
              call cadzow_group (obj, ierr)
              if ( ierr /= 0) then
                 ntr = FATAL_ERROR
                 write(obj%print_lun,*) ' FATAL_ERROR in cadzow_group '
                 return
              end if

              obj%iout = 0 
              obj%ioinpt = 0 
              obj%ntr_panel = 0

              if ( obj%ifirst) then
                obj%isout = 1 
                obj%ieout = obj%num_tr
              else 
                obj%isout = 1 + obj%num_pts
                obj%ieout = obj%num_tr+ obj%num_pts
              end if 
              obj%nout = obj%ieout - obj%isout + 1
              obj%ifirst = .false.
              obj%nread = obj%num_tr
                
            else 
              obj%hdval_old = obj%hdval  
              ntr = NEED_TRACES
              return  
            endif 
 
          else 

!..... check the case if number of input traces <  num_tr in the 
!           same gather

            iadd  = 0

            if (obj%itrseq >= (obj%num_tr + obj%num_pts) ) then
              if (obj%ntr_panel <= 0) then
                if (obj%ifirst) then
                  return 
                else if (obj%itrseq <= obj%num_tr + obj%num_pts) then
                  obj%ntr_panel = obj%num_tr + obj%num_pts
                  iadd = obj%num_tr
                   obj%ifirst = .true.
                else 
                 obj%ntr_panel = obj%num_tr + 2*obj%num_pts
                 iadd = obj%num_tr+obj%num_pts
                 obj%ifirst = .true.
                end if ! first
              end if  ! ntr_panel
            end if  ! itrseq
! 
            if (obj%hdval_old/=obj%hdval .and. ntr==0) obj%done = 1 
            if (ntr /= 0) then 
              obj%hd_save(1:obj%nwih) = hd(1:obj%nwih,ntr)
              obj%tr_save(1:obj%ndpt) = tr(1:obj%ndpt,ntr)
            endif 
 
!..... Apply FX Cadzow Filter if number of input traces > num_tr.
 
            ierr  = 0
            obj%ilast = .true. 
            if (obj%itrseq < (obj%num_tr + obj%num_pts) .and. obj%ifirst) then
              obj%outtr(1:obj%ndpt, 1:obj%ntr_panel) =      &
              obj%trstor(1:obj%ndpt, 1:obj%ntr_panel) 
              obj%outwt(1:obj%ndpt, 1:obj%ntr_panel) = 1.0 
              obj%indx_tr = (/(i, i=1,obj%nxstor)/)  ! implied loop
            else 

!.....         catch this special case

               if (obj%nxstor > obj%itrseq ) then
                 do ipos = obj%itrseq+1, obj%nxstor
                   obj%iseq(ipos) = ipos
                   obj%hdstor(1:obj%nwih, ipos) = 0.0 
                   obj%trstor(1:obj%ndpt, ipos) = 0.0 
                 end do
               end if	! nxstor
  
               call cadzow_group (obj, ierr)
            endif 

            if ( ierr /= 0) then
              ntr = FATAL_ERROR
                 write(obj%print_lun,*) ' FATAL_ERROR after cadzow_group'
              return
            end if ! ierr
            
            obj%iout = 0 
            obj%ioinpt = 0 

            if (obj%ifirst .and. obj%ilast) then        
              obj%isout = 1 + iadd
              obj%ieout = obj%ntr_panel 
            else 
              obj%isout = min(obj%itrseq,obj%nxstor) -    &
                          (obj%ntr_panel+obj%num_pts) + 1 
              obj%ieout = min(obj%itrseq,obj%nxstor) 
            end if ! first last
            obj%nout = obj%ieout - obj%isout + 1
            obj%ntr_panel = 0  
          endif 

      endif 
 
!..... ioinpt = 0 -- output filtered traces.
 
      if (obj%ioinpt == 0) then 
 
        if (obj%iout < obj%nout) then 
          ntr = 1
          call cadzow_out (obj, ntr, tr, hd)
          ! is this first trace of ensemble?
          if(NINT(hd(4,1))==1 .and. obj%svd_rank<5) then
! addback half of original trace
		   do i=1,obj%ndpt
		    tr(i,1)=(obj%trstor(i,1)+tr(i,1))*0.5
		   end do
		  endif
  935     format(' output cdp,isout,ieout ',3f8.1,2i6)
 
          
        else 
 
!.... start a new line or gather

          if (obj%hdval_old /= obj%hdval) then 

            obj%iseq = (/(i, i=1,obj%nxstor)/) ! implied loop
            obj%nread = obj%num_tr + obj%num_pts
            obj%ilast = .false.
            obj%ifirst = .true.

            obj%ntr_panel = 1 
            obj%itrseq = 1 
            ipos = mod(obj%itrseq,obj%nxstor) 
            if (ipos == 0) ipos = obj%nxstor 
            obj%iseq(ipos) = obj%itrseq 
!
            obj%hdstor(1:obj%nwih, ipos) = obj%hd_save(1:obj%nwih)
            obj%trstor(1:obj%ndpt, ipos) = obj%tr_save(1:obj%ndpt)
 
          endif ! hdval
 
          obj%ioinpt = 1 
          obj%nout = 0 
          obj%iout = 0 		! reset iout
 
          if (obj%done == 1) then 
            ntr = NO_MORE_TRACES 
            return  
          else 
            obj%hdval_old = obj%hdval 
            ntr = NEED_TRACES
            return  
          endif ! done
        endif 
      endif 
      return
      end subroutine cadzow_single_trace

 
      subroutine cadzow_group(obj, ierr)

!  purpose:   Filter a group of input data.

      implicit none
      type(cadzow_struct),intent(inout) :: obj          ! arguments
      integer, intent(inout)          :: ierr            ! arguments

      integer  :: nthpts, nfft                           ! local
      integer  :: ixst_in, ixend_in, itst, itend         ! local
      integer  :: ixst_out, ixend_out                    ! local
      integer  :: iw, ii, i, j                           ! local
      real     :: mute_min, rmute, rmsamp, scale         ! local
	real   :: tracest,tracend

      real     :: rw1fft(obj%nfftp2,obj%nxstor)          ! local
      complex  :: cw1fft(obj%nhftp1,obj%nxstor)          ! local

      real     :: work1(max(obj%nfftp2,obj%nxstor))      ! local
      complex  :: cwork(max(obj%nfftp2,obj%nxstor))      ! local
      real     :: seq(obj%nxstor)                        ! local

      ierr = 0

      nfft  =  obj%nfftp2 - 2
      scale = 1.0/(nfft) 

      seq(1:obj%nxstor) = obj%iseq(1:obj%nxstor)
      obj%indx_tr = (/(j, j=1,obj%nxstor)/)
 
      call sort_qkisort (obj%nxstor, seq, obj%indx_tr)

!.... set the output weight and buffer to zero.
 
      obj%outtr = 0.0 
      obj%outwt = 0.0

!.... Load a group of traces into buffer
! ixst_out=1 for first window
      if (obj%ifirst .and. obj%ilast) then 
        ixst_in = 1
        ixend_in = obj%ntr_panel
        ixst_out = 1
        ixend_out = obj%ntr_panel
      else
        ixst_in = 1
        if ( obj%ifirst) then 				! first group of gather
          ixend_in = obj%num_tr + obj%num_pts
          ixst_out = 1
          ixend_out = obj%num_tr
        else 
          ixend_in = obj%num_tr + 2*obj%num_pts
          ixst_out = obj%num_pts+1
          ixend_out = obj%num_pts+obj%num_tr
        end if
        if (obj%ilast) then
	 ixend_out = obj%num_tr + 2*obj%num_pts
	end if
      end if
! need num_pts even and num_tr odd
      itst = 1 
      itend = obj%ntpts
      nthpts = obj%ntpts/2 
      mute_min = 9999 
 
      rmute = minval(obj%hdstor(2,obj%indx_tr(ixst_in:ixend_in)))
      mute_min = min(rmute, mute_min) 
	tracest=obj%hdstor(4,obj%indx_tr(ixst_in))
	tracend=obj%hdstor(4,obj%indx_tr(ixend_in))
!.... Perform filter in a time window

       do iw = 1, obj%nwin 
          if (mute_min <= itend) then  
          rmsamp = 0.
          do j = ixst_in, ixend_in  
            work1(1:itend-itst+1) = obj%trstor(itst:itend,obj%indx_tr(j)) 
            rmsamp = rmsamp +     &
                    sum(work1(1:itend-itst+1)*work1(1:itend-itst+1))
            rmsamp = sqrt(rmsamp)  
 
!....... pad zeros to the data before calling fft
 
            work1(obj%ntpts+1:obj%nfftp2) = 0. 

!.... compute fft of windowed data.
 
            call fft_rc_transform(obj%fftrc_obj, work1, cwork) 
            cw1fft(1:obj%nhftp1,j) = cwork(1:obj%nhftp1)
 
          end do 
 
          if (rmsamp <= 1.e-10) then
           write(6,*) 'Skipping a group because rmsamp < 1.e-10'
           go to 100
          endif
! cw1fft is the f-x data for the roving window
          call cadzow_sol (obj, ixst_in, ixend_in, ixst_out, ixend_out, &
               cw1fft, ierr) 
          if ( ierr /= 0) then
              return
          end if
 
!... inverse FFT 
          do j = ixst_out, ixend_out

            call fft_cr_transform(obj%fftcr_obj, cw1fft(1,j), rw1fft(1,j), &
            scale)

            ii = 0 
            do i = itst, itend
              ii = ii + 1
              obj%outtr(i,j) = obj%outtr(i,j)+ rw1fft(ii,j)*obj%winwt(ii)
              obj%outwt(i,j) = obj%outwt(i,j)+ obj%winwt(ii)
            end do
          end do 
 
        endif 
  100   continue 
 
        itst = itst + nthpts 
        itend = itend + nthpts 
        if (itend <= obj%ndpt) cycle  
        itend = obj%ndpt 
        itst = obj%ndpt - obj%ntpts + 1 
      end do

      return  
      end subroutine cadzow_group

 
      subroutine cadzow_out(obj, ntr, tr, hd)
!  only called from cadzow_single_trace
!  purpose:   output filtered traces.
!
!  input:     trstor    real    buffer of a group of input traces.
!             hdstor    dble    buffer of a group of input trace headers.

      implicit none
      type(cadzow_struct),intent(inout) :: obj              ! arguments

      integer , intent(in)  :: ntr                           ! arguments
      real ,    intent(inout) :: tr(obj%ndpt,ntr)            ! output
      double precision , intent(inout) :: hd(obj%nwih,ntr)   ! output
 
      integer :: jtr, jhd                                    ! local
      real , allocatable :: tmp1u(:)                         ! local
 

      jhd = obj%indx_tr(obj%isout+obj%iout)
      jtr = obj%isout + obj%iout  
      obj%iout = obj%iout + 1

      allocate (tmp1u(obj%ndpt)) 
      tmp1u = obj%outwt(:,jtr) 
      where (tmp1u /= 0.)  
        tr(:,ntr) = obj%outtr(:,jtr)/tmp1u 
      elsewhere 
        tr(:,ntr) = obj%outtr(:,jtr) 
      end where 

      deallocate (tmp1u) 
 
      hd(1:obj%nwih,ntr) = obj%hdstor(1:obj%nwih,jhd) 
 
!..... restore front and end mute

      call mutehw (hd, tr, obj%ndpt, 0.0, ikill = MUTEHW_BOTH)
      call lav_set_hdr (hd, tr, obj%ndpt,ntr)
    
      return  
      end subroutine cadzow_out


      subroutine cadzow_sol(obj, ixst_in, ixend_in, ixst_out, ixend_out, &
            cw1fft, ierr)

!  purpose:   Compute qnd apply  filter for each frequency slice.

!
!  input:     ixst_in    int     starting location of input array.
!             ixend_in   int     ending location of input array.
!             ixst_out   int     starting location of output array.
!             ixend_out  int     ending location of output array.
!             cw1fft    cmplx    array containing fft of the window.

      implicit none

      type(cadzow_struct),intent(inout) :: obj                 ! arguments
      integer, intent(in)          :: ixst_in, ixend_in         ! arguments
      integer, intent(in)          :: ixst_out, ixend_out       ! arguments
      integer, intent(inout)       :: ierr                      ! arguments
                 
      complex , intent(inout) :: cw1fft(obj%nhftp1,obj%nxstor)  ! arguments
                                        
      integer :: i, i1, i2, ix, kf, i0, nv, nu, p, k, ns        ! local
      integer :: ndsign, ir1, ir2, m, j, n, mmax, nmax          ! local
      integer :: nfirst_freq
      integer :: navg(obj%nxstor)
      complex :: dotp
      complex  :: cdata1(obj%nxstor)                            ! frequency slice
      complex  :: cdata2(obj%nxstor)                            ! filtered slice
!      complex  :: cwork(2*obj%nxstor)                           ! local
!      complex  :: cwork2(2*obj%nxstor)                          ! local
      complex  :: A(39,39)			! Hankel
      complex  :: U(39,39)
      complex  :: V(39,39)
      complex  :: B(39,39)					! product
      real     :: S(39)						! singular

      real      :: taper_weight                                 ! local
      real      :: taper_len1, taper_len2, a0, a1, a2
       
!...... initialize parameters
 
      ierr = 0 
      ndsign = ixend_in-ixst_in+1	! must be odd >> CHECK
      m = ndsign/2+1
      n = m				! square matrix
!      mmax = m
!      nmax = m
      nv = m
      nu = m
      p = 0
      taper_len1 = obj%ifreq1 + 1
      taper_len2 = obj%nhftp1 - obj%ifreq2 + 1

      if (obj%keep_lo_freq_data) then
         nfirst_freq = obj%ifreq1
      else 
         nfirst_freq = 1
      end if 
! does higher frequencies, but tapers down
      do kf = nfirst_freq, obj%nhftp1  
 
!     compute frequency taper weight

         taper_weight = 1.0
         if ( kf < obj%ifreq1) then
           taper_weight = kf /taper_len1
         else if ( kf  > obj%ifreq2 ) then
           taper_weight = (obj%nhftp1-kf)/taper_len2
         end if

         taper_weight = taper_weight*taper_weight
! cdata1 is freq slice
        cdata1(ixst_in:ixend_in) = cw1fft(kf,ixst_in:ixend_in)
! example n=7 m=4 => 4 row 4 col
! 1 2 3 4
! 2 3 4 5
! 3 4 5 6
! 4 5 6 7
!   build Hankel trajectory matrix (row,col)
!	A(1,1)=cdata1(ixst_in)
!	A(1:39,1:39)=cmplx(0.0)
!	U(1:39,1:39)=cmplx(0.0)
!	V(1:39,1:39)=cmplx(0.0)
	do i=1,m
	 do j=1,m
          A(i,j)=cdata1(j+ixst_in-2+i)
	 end do
	end do
	B(1:m,1:m)=cmplx(0.0)
! only compute as many columns as needed
!                   b  i  i  i  i             i             o  o  o
        call svdc ( A, m, n, p, obj%svd_rank, obj%svd_rank, S, U, V )
! multiply left matrix U by a few bits of S(diagonal)
! right columns of result will be zero
	do i=1,obj%svd_rank		! reduce matrix rank
	 do j=1,m	! row
	  B(j,i)=S(i)*U(j,i)	! 2nd index of B only up to rank
	 end do
	end do	! i
! multiply by conjugate transpose V
	do j=1,m		! use all for j?
	 do i=1,m
	  dotp=cmplx(0.0)
	  do k=1,obj%svd_rank		!
	   dotp=dotp+B(i,k)*conjg(V(j,k))
	  end do
	  A(i,j)=dotp
	 end do	! i
	end do	! j
	cdata2(1:)=cmplx(0.0)
	navg(1:)=0
! average the Hankel diagonals
	do i=1,m
	 do j=1,m
	  k=j+ixst_in-2+i
          cdata2(k)=cdata2(k)+A(i,j)
	  navg(k)=navg(k)+1
	 end do
	end do	! i
	do i=ixst_in,ixend_in
	 cdata2(i)=cdata2(i)/real(navg(i))
	end do

         cw1fft(kf,ixst_out:ixend_out) = cdata2(ixst_out:ixend_out) 
           
      end do ! kf
 
      return  
      end subroutine cadzow_sol

!

!</execute_only>


!!------------------------------- wrapup -----------------------------------!!
!!------------------------------- wrapup -----------------------------------!!
!!------------------------------- wrapup -----------------------------------!!


!<execute_only>

      subroutine cadzow_wrapup (obj)
      implicit none
      type(cadzow_struct),intent(inout) :: obj       ! arguments

      if (obj%skip_wrapup) return
      obj%skip_wrapup = .true.

      call singletracesupport_delete (obj%sts)
      return
      end subroutine cadzow_wrapup

!</execute_only>
!                       b  i  i  i  i   i   o  o  o
      subroutine svdc ( a, m, n, p, nu, nv, s, u, v )
! Modified slightly for Fortran 90
!*********************************************************************72
!c CSVD computes the singular value decomposition of an M by N complex matrix.
!  Discussion:
!    This routine requires that N <= M.
!    The singular value decomposition of a complex M by N matrix A
!    has the form
!      A = U S V*
!    where
!      U is an M by M unitary matrix,
!      S is an M by N diagonal matrix,
!      V is an N by N unitary matrix.
!    Moreover, the entries of S are nonnegative and occur on the diagonal
!    in descending order.
!    Several internal arrays are dimensioned under the assumption
!    that N <= 100.
!    Thanks to Vladimir Rutsky for pointing out, on 10 January 2013,
!    that the line
!      IF ( N .LT. 2 ) GO TO 570
!    had been incorrectly modified to
!      if ( 2 .lt. n ) then
!    when
!      if ( 1 .lt. n ) then
!    was required.
!  Modified:
!    10 January 2013
!  Author:
!    Peter Businger, Gene Golub.
!    Modifications by John Burkardt.
!  Reference:
!    Peter Businger, Gene Golub,
!    Algorithm 358:
!    Singular Value Decomposition of a Complex Matrix,
!    Communications of the ACM,
!    Volume 12, Number 10, October 1969, pages 564-565.
!  Parameters:
!    Input/output, complex A(MMAX,*), the M by N matrix, which may be
!    augmented by P extra columns to which the transformation U*
!    is to be applied.  On output, A has been overwritten, and
!    if 0 < P, columns N+1 through N+P have been premultiplied by U*.
!
!    Input, integer M, N, the number of rows and columns in A.
!    It must be the case that 1 <= N <= M.  Several internal arrays are
!    dimensioned under the assumption that N <= NBIG, where NBIG
!    is an internal parameter, currently set to 100.
!
!    Input, integer P, the number of vectors, stored in A(*,N+1:N+P),
!    to which the transformation U* should be applied.
!
!    Input, integer NU, the number of columns of U to compute.
!
!    Input, integer NV, the number of columns of V to compute.
!
!    Output, real S(N), the computed singular values.
!
!    Output, complex U(MMAX,NU), the first NU columns of U.
!
!    Output, complex V(NMAX,NV), the first NV columns of V.
!
!  Local Parameters:
!
!    Local, real ETA, the relative machine precision.
!    The original text uses ETA = 1.5E-8.
!
!    Local, integer NBIG, is a parameter used to dimension work arrays.
!    The size of NBIG limits the maximum possible size of N that can
!    be handled.  If you want to work with values of N that are larger,
!    simply increase the value assigned to NBIG below.
!
!    Local, real TOL, the smallest normalized positive number, divided by ETA.
!    The original test uses TOL = 1.E-31.
!
      implicit none

      integer,intent(in) :: m
      integer,intent(in) :: n
      integer,intent(in) :: nu
      integer,intent(in) :: nv
      integer,intent(in) :: p
      real,   intent(out) :: s(:)
      complex,intent(out) :: u(:,:)
      complex,intent(out) :: v(:,:)
      complex,intent(inout) :: a(:,:)

      integer :: nbig
      parameter ( nbig = 40 )

      real :: b(nbig)
      real :: c(nbig)
      real :: t(nbig)
      real :: cs,eps,eta,f,g,h
      real :: sn,tol,w,x,y,z
      integer :: i,j,k,kk,k1,l,l1,ll
      complex :: q,r

!      save eta
!      save tol

      data eta / 1.1920929e-07 /
      data tol / 1.5e-31 /
!
!  Check N.
!
      if ( n < 1 ) then
        write ( *, '(a)' ) 'CSVD - Fatal error!'
        write ( *, '(a)' ) '  Input N < 1.'
        stop
      else if ( nbig < n ) then
        write ( *, '(a)' ) 'CSVD - Fatal error!'
        write ( *, '(a)' ) '  NBIG < N.'
        stop
      end if
!
!  Check M.
!
      if ( m < 1 ) then
        write ( *, '(a)' ) 'CSVD - Fatal error!'
        write ( *, '(a)' ) '  Input M < 1.'
        stop
      else if ( m < n ) then
        write ( *, '(a)' ) 'CSVD - Fatal error!'
        write ( *, '(a)' ) '  M < N.'
        stop
      end if
!
!  Householder reduction.
! U and V are zeroed before use
!
      c(1:m) = 0.0e+00
      b(1:m) = 0.0e+00
      t(1:m) = 0.0e+00
      k = 1

10    continue

      k1 = k + 1
!
!  Elimination of A(I,K), I = K+1, ..., M.
!
      z = 0.0e+00
      do i = k, m
        z = z + ( real ( a(i,k) ) )**2 + ( aimag ( a(i,k) ) )**2
      end do

      b(k) = 0.0e+00

      if ( tol .lt. z ) then

        z = sqrt ( z )
        b(k) = z
        w = cabs ( a(k,k) )

        if ( w .eq. 0.0e+00 ) then
          q = cmplx ( 1.0e+00, 0.0e+00 )
        else
          q = a(k,k) / w
        end if

        a(k,k) = q * ( z + w )

        if ( k .ne. n + p ) then
          do j = k1, n + p
            q = cmplx ( 0.0e+00, 0.0e+00 )
            do i = k, m
              q = q + conjg ( a(i,k) ) * a(i,j)
            end do
            q = q / ( z * ( z + w ) )
            do i = k, m
              a(i,j) = a(i,j) - q * a(i,k)
            end do
          end do
!
!  Phase transformation.
!
          q = -conjg ( a(k,k) ) / cabs ( a(k,k) )

          do j = k1, n + p
            a(k,j) = q * a(k,j)
          end do

        end if

      end if
!
!  Elimination of A(K,J), J = K+2, ..., N
!
      if ( k .eq. n ) then
        go to 140
      end if

      z = 0.0e+00
      do j = k1, n
        z = z + ( real ( a(k,j) ) )**2 + ( aimag ( a(k,j) ) )**2
      end do

      c(k1) = 0.0e+00

      if ( tol .lt. z ) then

        z = sqrt ( z )
        c(k1) = z
        w = cabs ( a(k,k1) )

        if ( w .eq. 0.0e+00 ) then
          q = cmplx ( 1.0e+00, 0.0e+00 )
        else
          q = a(k,k1) / w
        end if

        a(k,k1) = q * ( z + w )

        do i = k1, m
          q = cmplx ( 0.0e+00, 0.0e+00 )
          do j = k1, n
            q = q + conjg ( a(k,j) ) * a(i,j)
          end do
          q = q / ( z * ( z + w ) )
          do j = k1, n
            a(i,j) = a(i,j) - q * a(k,j)
          end do
        end do
!
!  Phase transformation.
!
        q = -conjg ( a(k,k1) ) / cabs ( a(k,k1) )
        do i = k1, m
          a(i,k1) = a(i,k1) * q
        end do

      end if

      k = k1
      go to 10
!
!  Tolerance for negligible elements.
!
140   continue

      eps = 0.0e+00
      do k = 1, n
        s(k) = b(k)
        t(k) = c(k)
        eps = amax1 ( eps, s(k) + t(k) )
      end do

      eps = eps * eta
!
!  Initialization of U and V.
!
      if ( 0 .lt. nu ) then

        do j = 1, nu
          do i = 1, m
            u(i,j) = cmplx ( 0.0e+00, 0.0e+00 )
          end do
          u(j,j) = cmplx ( 1.0e+00, 0.0e+00 )
        end do

      end if

      if ( 0 .lt. nv ) then
        do j = 1, nv
          do i = 1, n
            v(i,j) = cmplx ( 0.0e+00, 0.0e+00 )
          end do
          v(j,j) = cmplx ( 1.0e+00, 0.0e+00 )
        end do
      end if
!
!  QR diagonalization.
!
      do kk = 1, n

        k = n + 1 - kk
!
!  Test for split.
!
220     continue

        do ll = 1, k
          l = k + 1 - ll
          if ( abs ( t(l) ) .le. eps ) then
            go to 290
          end if
          if ( abs ( s(l-1) ) .le. eps ) then
            go to 240
          end if
        end do
!
!  Cancellation of E(L).
!
240     continue

        cs = 0.0e+00
        sn = 1.0e+00
        l1 = l - 1

        do i = l, k

          f = sn * t(i)
          t(i) = cs * t(i)

          if ( abs ( f ) .le. eps ) then
            go to 290
          end if

          h = s(i)
          w = sqrt ( f * f + h * h )
          s(i) = w
          cs = h / w
          sn = - f / w

          if ( 0 .lt. nu ) then
            do j = 1, n
              x = real ( u(j,l1) )
              y = real ( u(j,i) )
              u(j,l1) = cmplx ( x * cs + y * sn, 0.0e+00 )
              u(j,i)  = cmplx ( y * cs - x * sn, 0.0e+00 )
            end do
          end if

          if ( p .ne. 0 ) then
            do j = n + 1, n + p
              q = a(l1,j)
              r = a(i,j)
              a(l1,j) = q * cs + r * sn
              a(i,j)  = r * cs - q * sn
            end do
          end if

        end do
!
!  Test for convergence.
!
290     continue

        w = s(k)

        if ( l .eq. k ) then
          go to 360
        end if
!
!  Origin shift.
!
        x = s(l)
        y = s(k-1)
        g = t(k-1)
        h = t(k)
        f = ((y - w)*(y + w)+(g - h)*(g + h))/(2.0e+00*h*y )
        g = sqrt ( f * f + 1.0e+00 )
        if ( f .lt. 0.0e+00 ) then
          g = -g
        end if
        f = ( ( x - w ) * ( x + w ) + ( y / ( f + g ) - h ) * h ) / x
!
!  QR step.
!
        cs = 1.0e+00
        sn = 1.0e+00
        l1 = l + 1

        do i = l1, k

          g = t(i)
          y = s(i)
          h = sn * g
          g = cs * g
          w = sqrt ( h * h + f * f )
          t(i-1) = w
          cs = f / w
          sn = h / w
          f = x * cs + g * sn
          g = g * cs - x * sn
          h = y * sn
          y = y * cs

          if ( 0 .lt. nv ) then
            do j = 1, n
              x = real ( v(j,i-1) )
              w = real ( v(j,i) )
              v(j,i-1) = cmplx ( x * cs + w * sn, 0.0e+00 )
              v(j,i)   = cmplx ( w * cs - x * sn, 0.0e+00 )
            end do
          end if

          w = sqrt ( h * h + f * f )
          s(i-1) = w
          cs = f / w
          sn = h / w
          f = cs * g + sn * y
          x = cs * y - sn * g

          if ( 0 .lt. nu ) then
            do j = 1, n
              y = real ( u(j,i-1) )
              w = real ( u(j,i) )
              u(j,i-1) = cmplx ( y * cs + w * sn, 0.0e+00 )
              u(j,i)   = cmplx ( w * cs - y * sn, 0.0e+00 )
            end do
          end if

          if ( p .ne. 0 ) then
            do j = n + 1, n + p
              q = a(i-1,j)
              r = a(i,j)
              a(i-1,j) = q * cs + r * sn
              a(i,j)   = r * cs - q * sn
            end do
          end if

        end do

        t(l) = 0.0e+00
        t(k) = f
        s(k) = x
        go to 220
!
!  Convergence.
!
360     continue

        if ( w .lt. 0.0e+00 ) then
          s(k) = - w
          if ( 0 .lt. nv ) then
            do j = 1, n
              v(j,k) = - v(j,k)
            end do
          end if
        end if
      end do
!
!  Sort the singular values.
!
      do k = 1, n

        g = - 1.0e+00
        j = k

        do i = k, n
          if ( g .lt. s(i) ) then
            g = s(i)
            j = i
          end if
        end do

        if ( j .ne. k ) then

          s(j) = s(k)
          s(k) = g
!
!  Interchange V(1:N,J) and V(1:N,K).
!
          if ( 0 .lt. nv ) then
            do i = 1, n
              q      = v(i,j)
              v(i,j) = v(i,k)
              v(i,k) = q
            end do
          end if
!
!  Interchange U(1:N,J) and U(1:N,K).
!
          if ( 0 .lt. nu ) then
            do i = 1, n
              q      = u(i,j)
              u(i,j) = u(i,k)
              u(i,k) = q
            end do
          end if
!
!  Interchange A(J,N1:NP) and A(K,N1:NP).
!
          if ( p .ne. 0 ) then
            do i = n + 1, n + p
              q      = a(j,i)
              a(j,i) = a(k,i)
              a(k,i) = q
            end do
          end if
        end if
      end do
!
!  Back transformation.
!
      if ( 0 .lt. nu ) then
        do kk = 1, n
          k = n + 1 - kk
          if ( b(k) .ne. 0.0e+00 ) then
            q = -a(k,k) / cabs ( a(k,k) )
            do j = 1, nu
              u(k,j) = q * u(k,j)
            end do
            do j = 1, nu
              q = cmplx ( 0.0e+00, 0.0e+00 )
              do i = k, m
                q = q + conjg ( a(i,k) ) * u(i,j)
              end do
              q = q / ( cabs ( a(k,k) ) * b(k) )
              do i = k, m
                u(i,j) = u(i,j) - q * a(i,k)
              end do
            end do
          end if
        end do
      end if

      if ( 0 .lt. nv ) then
        if ( 1 .lt. n ) then
          do kk = 2, n
            k = n + 1 - kk
            k1 = k + 1
            if ( c(k1) .ne. 0.0e+00 ) then
              q = -conjg ( a(k,k1) ) / cabs ( a(k,k1) )
              do j = 1, nv
                v(k1,j) = q * v(k1,j)
              end do
              do j = 1, nv
                q = cmplx ( 0.0e+00, 0.0e+00 )
                do i = k1, n
                  q = q + a(k,i) * v(i,j)
                end do
                q = q / ( cabs ( a(k,k1) ) * c(k1) )
                do i = k1, n
                  v(i,j) = v(i,j) - q * conjg ( a(k,i) )
                end do
              end do
            end if
          end do
        end if
      end if

      return
      end subroutine

!!
!!
!!----------------------------- end of module ------------------------------!!
!!----------------------------- end of module ------------------------------!!
!!----------------------------- end of module ------------------------------!!


      end module cadzow_module


!!--------------------------------- end ------------------------------------!!
!!--------------------------------- end ------------------------------------!!
!!--------------------------------- end ------------------------------------!!


