!<CPS_v1 type="PROCESS"/>
!!------------------------------ kpstm2d.f90 -------------------------------!!
!!------------------------------ kpstm2d.f90 -------------------------------!!
!!------------------------------ kpstm2d.f90 -------------------------------!!

!<license>
!-------------------------------------------------------------------------------
! Copyright (c) 2007 ConocoPhillips Company
!
! Permission is hereby granted, free of charge, to any person obtaining a copy
! of this software and associated documentation files (the "Software"), to deal
! in the Software without restriction, including without limitation the rights
! to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
! copies of the Software, and to permit persons to whom the Software is
! furnished to do so, subject to the following conditions:
!
! The above copyright notice and this permission notice shall be included in all
! copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
! AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
! SOFTWARE.
!-------------------------------------------------------------------------------
!</license>

!<brief_doc>
!-------------------------------------------------------------------------------
!                         C P S   P R O C E S S             
!
! Name       : KPSTM2D	Kirchhoff time migration 2D
! Category   : Migration
! Written    : 2018-4-4   by: Michael Seman
! Revised    : 
!
! Maturity   : production
! Purpose    : straight-ray pre-stack migration in 2D
! Portability: No known limitations.
! Parallel   : Yes
!
!-------------------------------------------------------------------------------
!</brief_doc>

!<descript_doc>
!-------------------------------------------------------------------------------
!                          GENERAL DESCRIPTION              
! Performs PSTM by Kirchhoff double square root method with antialiasing.
! Suitable only for simple geology: small lateral velocity variation,
! isotropic and no out-of-plane events. If your data shows non-hyperbolic
! moveout, then PSTM will not be wonderful. It only deals with primary
! reflections and diffractions. Other entities such as P-S converted waves
! will not be handled correctly.
! Anti-alias filtering is zero-phase (note if you did minimum-phase jobflow).
!-------------------------------------------------------------------------------
!</descript_doc>

!<advice_doc>
!-------------------------------------------------------------------------------
!                            ADVICE FOR USERS
! Common-offset gathers must be regular. Use FILL process to pad missing traces
! Input data must be pre-processed, e.g. noise spikes removed and multiples
! attenuated and static corrections applied (land).
! RMS velocities must match the CDP range of traces (HDR 7 and 8).
! If shallow data is the target, maximum angle should be limited to 55 degrees.
! Also note the straight-ray formula may become less veridical as the
! maximum angle is pushed ever higher.
! After sorting migration back to CDPs, it is likely you will need to pick a
! new mute.
!    Data preparation:
! no normal moveout
! mute refractions.
! regularise offsets
! avoid low-cut filtering in pre-processing
! no AGC, but surface-consistent amplitude balance instead
! Gain of T to power 0.5 is correct for 2D migration, not spherical
! divergence (TV**2). After migration, further gain may be applied to boost
! boost deep events to personal tastes.
!  To compare with unmigrated stack generated from same data, different gains
! will be needed. For example if unmigrated stack had TPOW t**N, suggest for
! migrated stack to use TPOW t**N-0.5
! When picking velocities in va, use control panel to show previous and next
! locations. Velocities to use for migration may need to be smoother than
! what yields best unmigrated stack. For velocity revision, apply inverse NMO
! to the migrated gathers, and repeat SVA velocity analysis. Using DMO for
! the initial velocity analysis (before migration) may help.
! If gathers are nearly flat after migration, then residual NMO from VTRIM
! may suffice to finish up.
!    Usually for migration, the "common-offset" gathers are binned offsets.
! So each bin contains 2 or more offsets. This blurs the migration unless the
! offsets are regularised (by partial NMO for example). An alternative is to
! sort by recording channel, so the offsets are uniform. Then for parameter
! CDPINC enter shot interval. For shot interval of 25 m or less, this method
! is satisfactory.
!  To minimise the high-cut effects of the anti-aliasing filter, the input
! trace gathers may be interpolated, which reduces aliasing. However, this
! doubles the number of traces, and obviously increases the runtime.
!   This process may be run parallel. For testing, a few thousand CDPs may
! be done on a single CPU (icps).
!
! For impulse response test:
! Suggest to create a series of spikes 200 to 300 ms apart and filter them with
! a low-pass filter of say 40 Hz. If you use a single-sample spike, the
! anti-alias filter will almost kill it at steep angles.
!-------------------------------------------------------------------------------
!</advice_doc>

!<trace_in_doc>
!-------------------------------------------------------------------------------
!                      TRACE INPUT REQUIREMENTS                
!
! Process is a gather-trace process.
!
! Traces must be sorted into common offsets (header 6 or 10).
!
!-------------------------------------------------------------------------------
!</trace_in_doc>

!<trace_out_doc>
!-------------------------------------------------------------------------------
!                     TRACE OUTPUT CHARACTERISTICS             
!
! This process does alter input traces.
! This process may alter input headers.
! This process outputs the same traces as it receives, possibly with altered
! headers.
!
! This process outputs traces with same gather status as the input traces.
!
!-------------------------------------------------------------------------------
!</trace_out_doc>

!<global_doc>
!-------------------------------------------------------------------------------
!       PROJECT DATA, JOB DATA, AND GLOBAL PARAMETERS USED OR CHANGED      
!
! Name      Description                             Action taken
! ----      -----------                             ------------
! NWIH      number of words in trace header         used but not changed
! DT		sample interval	
! NDPT		number of samples
! GATHER	gathered status							tested
!
!
!-------------------------------------------------------------------------------
!</global_doc>

!<header_word_doc>
!-------------------------------------------------------------------------------
!                   TRACE HEADER WORDS USED OR CHANGED           
! 
! Hwd#    Description                Action taken
! ----    -----------                ------------
! 6	  	offset				used
! 2		top mute			used
! 25		LAV				recalculated
!-------------------------------------------------------------------------------
!</header_word_doc>

!<history_doc>
!-------------------------------------------------------------------------------
!                             REVISION HISTORY                    
! 
!    Date        Author      Description
!    ----        ------      -----------
! 1. 2018-4-4  Michael   Initial version
!
!-------------------------------------------------------------------------------
!</history_doc>

!<portability_doc>
!-------------------------------------------------------------------------------
!                        PORTABILITY LIMITATIONS               
!
! 64-bit build strongly recommended
!
!-------------------------------------------------------------------------------
!</portability_doc>

!<compile_doc>
!-------------------------------------------------------------------------------
!                      SPECIAL COMPILING REQUIREMENTS             
!
! Intel or GNU compilers.
! 
!-------------------------------------------------------------------------------
!</compile_doc>

!<calling_doc>
!-------------------------------------------------------------------------------
!                    SPECIFIC CALLING CHARACTERISTICS       
! Control
! Parameter     Value
! Name          Reported   Description
! ----          --------   -----------
! NTAPES           0       number of magnetic tapes needed.
! NEED_REQUEST   false     whether this process ever needs to request traces.
! NEED_LABEL     false     whether this process needs a label.     
! TWOSETS        false     whether this process needs two trace/header arrays.
! NSCRATCH         0       amount of temporary memory needed.       
! NSTORE           0       amount of permanent memory needed.      
! IFTD           false     whether this process frees tape drives. 
! NDISK            0       disk space needed (megabytes) if large. 
! SETUP_ONLY     false     whether this process is setup-only.    
!
!
! Upon input, NTR must have one of these values:
!  NTR >= 1              means to process the input traces.
!  NTR == NO_MORE_TRACES means there are no more input traces.
!
! Upon output, NTR will have one of these values:
!  NTR >= 1              if this process is outputting traces.
!  NTR == NO_MORE_TRACES if there are no more traces to output.
!  NTR == FATAL_ERROR    if this process has a fatal error.
!
!-------------------------------------------------------------------------------
!</calling_doc>

!<int_calling_doc>
!-------------------------------------------------------------------------------
!                    ALTERNATE INTERNAL CALLING METHODS          
!
!  None provided.
!
!-------------------------------------------------------------------------------
!</int_calling_doc>

!<algorithm_doc>
!-------------------------------------------------------------------------------
!                   ALGORITHM DESCRIPTION FOR DEVELOPERS         
!
! Performs basic Kirchhoff double square root migration, based on Schneider, W.
! (1978, Geophysics) and Sattlegger et al. (1980, Geophysical Prospecting).
! Anti-aliasing is by the Gray, S. (1992, Geophysical Prospecting) method of
! lowpass filtering according to the angle.  19 filters are used, plus the
! unfiltered trace.  A weighted sum of nearest filters is used.
! 45 degree phase shift is optionally applied at start of process.
! Fractional interpolation done by cubic splines.
! Angle taper is 10% of maximum angle plus 2 degrees. So for 80 degrees, taper
! is 10 degrees. Kirchhoff migration can theoretically do 90 degrees, so 80 is
! the maximum user can specify.
!-------------------------------------------------------------------------------
!</algorithm_doc>

!<programming_doc>
!-------------------------------------------------------------------------------
!                            PROGRAMMING NOTES                 
!
! Arrays are allocated after a getting gather. Memory use will be more than
! 2.5x the size of the input offset gather. Velocities array is 16-bit integer
! to save memory.
! Hence a limit of 32700 traces and 4000 samples is imposed. No temporary file
! is used, as workstations now have more RAM than when CPSeis was conceived.
! For 2D migration, runtime is proportional to:
! (number of CDPs)*(number of samples)*aperture
! since aperture increases with depth, runtime approximately depends on square
! of trace length.
! Setting DEBUG_LEVEL in JOB DATA will produce some extra printout in the job
! report. This is not recommended if running a parallel job.
!-------------------------------------------------------------------------------
!</programming_doc>

!-------------------------------------------------------------------------------
!
!<gui_def>
!<NS kpstm2d Process/NC=80/NR=40>
!
!              kpstm2d  Kirchhoff time migration 2D
!
!
!  PHASE45 = `CCCC    [/L] Phase shift 45 degree correction for 2D
!
!   MAXANG = `FFFF    [/L] Maximum angle to migrate (degrees)
!
!   VEL_PC = `FFFF    [/L] Percent of velocity for migration
!
!   CDPINC = `FFFF    [/L] Distance between CDP bins (feet or metres)
!
!  APERTUR = `FFFF    [/L] Aperture limit as distance
!
!Select VEL_PATHNAME[VEL_PATHNAME]`QSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
!
!                [VEL_PATHNAME_INFO]`XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
!
!<PARMS VEL_PATHNAME[/ML=128/XST]>
!<PARMS VEL_PATHNAME_INFO[/ML=128/XST]>
!
!</gui_def>
!1234567890123456789012345678901234567890123456789012345678901234567890123456789
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
!
!<HelpSection>
!
!<Help KEYWORD="PHASE45">
!<Tip> Rotate phase 45 degrees </Tip>
! Default = YES
! Allowed = NO
! Allowed = YES
! 45 degree phase for "half derivative" correction in 2D. Say YES unless this
! was done earlier in processing.
!</Help>
!
!<Help KEYWORD="MAXANG">
!<Tip> Maximum dip angle to migrate. </Tip>
! Default = 40
! Allowed = 1 - 80
! Maximum angle to migrate (at full amplitude). Amplitude is tapered down over
! the next 2 to 10 degrees. Steeper angle will increase runtime. Parameter
! APERTUR must be very large if steep angle is desired at deep times.
! Increasing MAXANG will result in more noise, when spatial sampling is
! coarse.
!</Help>
!
!<Help KEYWORD="VEL_PC">
!<Tip> Velocity percent multiplier for migration </Tip>
! Default = 100
! Allowed = 90-110
! RMS velocity field will be scaled by this percentage.
!</Help>
!
!<Help KEYWORD="CDPINC">
!<Tip> Distance between CMP bins </Tip>
! Default = 25
! Allowed = 1-999
! Calculations use this regular increment rather than co-ordinates in
! trace headers. Feet or metres as per survey. This will be shot interval if
! traces are sorted by recording channel rather than offset bin.
!</Help>
!
!<Help KEYWORD="APERTUR">
!<Tip> Maximum half aperture in feet or metres </Tip>
! Default = 3000
! Allowed = 250 - 9999
! Aperture is calculated from dip and velocity, but user may also
! further limit this as an absolute distance. Note this refers to one side
! so total aperture is twice this value. This is a hard limit (no taper).
! At the end time, the maximum angle available will be:
! arctan(APERTUR/(t*v))
! If APERTUR is too small, then Kirchhoff migration will produce horizontal
! events at deeper times.
!</Help>
!
!<Help KEYWORD="SELECT_VEL_PATHNAME">
!<Tip> Choose kpstm2d_PATHNAME using a file selection dialog box. </Tip>
!</Help>
!
!<Help KEYWORD="VEL_PATHNAME_INFO" TYPE= "DISPLAY_ONLY">
!<Tip> Status of VEL_PATHNAME. </Tip>
!</Help>
!
!<Help KEYWORD="VEL_PATHNAME">
!<Tip> Name for the velocity (RMS) file. </Tip>
! Default = NONE
! Allowed = char
! Velocity file must match CDP range and trace length.
!</Help>
!
!</HelpSection>
!
!-------------------------------------------------------------------------------

!!--------------------------- start of module ------------------------------!!
!!--------------------------- start of module ------------------------------!!
!!--------------------------- start of module ------------------------------!!

module kpstm2d_module
  !
  ! - Module linkages
  !
  use pc_module
  use getlun_module
  use named_constants_module
  use pathcheck_module
  use pathchoose_module
  use string_module
  use fft_module
  use nmo_module
  !
  implicit none
  !
  private
  !
  public :: kpstm2d_create
  public :: kpstm2d_initialize
  public :: kpstm2d_update
  public :: kpstm2d_delete
  public :: kpstm2d            ! main execution (trace processing) routine.
  public :: kpstm2d_wrapup


  character(len=100),public,save :: kpstm2d_IDENT = &
    '$Id: kpstm2d.f90,v 1.1 2018/4/17 13:45:44 Michael prod kpstm2d $'

  !!------------------------ parameter structure ---------------------------!!
  !!------------------------ parameter structure ---------------------------!!
  !!------------------------ parameter structure ---------------------------!!

  type, public :: kpstm2d_struct
    !
    private
    !
    logical                    :: skip_wrapup      ! wrapup flag.
    !
    ! - Process parameters
    !      Available from the parameter cache
    character(len=3)			:: phase45
    character(len=3)			:: unalias
    real				:: maxang
    real				:: vel_pc
    real				:: cdpinc
    real				:: apertur
    character(len=FILENAME_LENGTH)      :: vel_pathname
    type(pathchoose_struct),pointer     :: select_pathvel
    !
    integer				:: debug
    integer				:: plun
    integer				:: nclp			! number of filters
    integer				:: nfft			! FFT size
    real				:: freqny		! Nyquist
    real				:: minp			! minimum p
    real				:: degtap		! degrees for taper
    ! globals - assume TSTRT is zero for now....
    real				:: tstrt
    real				:: dt
    integer				:: ndpt
    ! - Statistics
    integer             :: trace_count
    integer				:: gc
    ! FFT
    type(fft_struct),pointer :: fftrc_obj           ! FFT
    type(fft_struct),pointer :: fftcr_obj           !
    type(nmo_struct),pointer :: nmov		! nmo vels
    !
  end type kpstm2d_struct


!!---------------------------- interfaces ---------------------------------!!
!!---------------------------- interfaces ---------------------------------!!
!!---------------------------- interfaces ---------------------------------!!

!!--------------------------------- data -----------------------------------!!
!!--------------------------------- data -----------------------------------!!
!!--------------------------------- data -----------------------------------!!

  type(kpstm2d_struct),pointer,save :: object      ! needed for traps.
! 32700x6.25 m = 204 km, should cope with most
      integer,parameter      :: DIMX = 32700	! maximum array sizes 
      integer,parameter      :: rev_noptions = 2
      character(len=6),save  :: rev_options(rev_noptions)
      data rev_options/'NO ', 'YES'/

contains

  !!--------------------------- kpstm2d_create -----------------------------!!
  !!--------------------------- kpstm2d_create -----------------------------!!
  !!--------------------------- kpstm2d_create -----------------------------!!

  subroutine kpstm2d_create (obj)
    !
    ! - Arguments
    !
    type (kpstm2d_struct), pointer :: obj
    !
    ! - Begin kpstm2d_create
    !
    allocate (obj)
    nullify  (obj%fftrc_obj) ! FFT
    nullify  (obj%fftcr_obj)
    nullify (obj%nmov)  
    nullify (obj%select_pathvel) 	! velocities
    ! default extension ?
    call pathchoose_create (obj%select_pathvel, 'vel_pathname', 'vel')
    !
    call kpstm2d_initialize (obj)
    !
  end subroutine kpstm2d_create

  !!--------------------------- kpstm2d_delete -----------------------------!!
  !!--------------------------- kpstm2d_delete -----------------------------!!
  !!--------------------------- kpstm2d_delete -----------------------------!!

  subroutine kpstm2d_delete (obj)
    !
    ! - Arguments
    !
    type (kpstm2d_struct), pointer :: obj
    !
    ! - Begin kpstm2d_delete
    !
    call kpstm2d_wrapup (obj)
    ! fft stuff
    if (associated(obj%fftrc_obj))  call fft_delete (obj%fftrc_obj)
    if (associated(obj%fftcr_obj))  call fft_delete (obj%fftcr_obj)
    ! test if associated?
    if (associated(obj%select_pathvel)) call pathchoose_delete (obj%select_pathvel)
    if (associated(obj%nmov)) call nmo_delete(obj%nmov)
    !
    deallocate(obj)
    !
  end subroutine kpstm2d_delete

  !!------------------------- kpstm2d_initialize ---------------------------!!
  !!------------------------- kpstm2d_initialize ---------------------------!!
  !!------------------------- kpstm2d_initialize ---------------------------!!

  subroutine kpstm2d_initialize (obj)
    !
    ! - Arguments
    !
    type(kpstm2d_struct),intent(inout) :: obj       ! arguments
    !
    ! - Begin kpstm2d_initialize
    !
	obj%maxang    = 40.0
	obj%vel_pc    = 100.0
	obj%cdpinc 	  = 25.0
	obj%apertur 	= 3000.0
	obj%trace_count = 0
	obj%gc 			= 0
	obj%vel_pathname          = PATHCHECK_EMPTY
	obj%plun = pc_get_lun()
    obj%phase45     = 'YES'
    obj%unalias		= 'YES'
    obj%nclp = 20
    !
    call kpstm2d_update (obj)
    !
  end subroutine kpstm2d_initialize

!!------------------------- start of update --------------------------------!!
!!------------------------- start of update --------------------------------!!
!!------------------------- start of update --------------------------------!!

  subroutine kpstm2d_update (obj)
    !
    ! - Arguments
    !
    type (kpstm2d_struct), intent (inout), target :: obj
    ! 
    ! - Local variables
    logical	:: gather
    integer	:: ifft,us,ier,stat,ma,mt,badinp
    real	:: minp,lowest
    !
    ! - Begin kpstm2d_update
    !
      object => obj                 ! needed for traps.
      obj%skip_wrapup = .true.    ! needed for the wrapup routine.

    !!------------------------ read parameters -----------------------------!!
    !!------------------------ read parameters -----------------------------!!
    !!------------------------ read parameters -----------------------------!!

    if (pathchoose_update(obj%select_pathvel, obj%vel_pathname)) return
    call pc_get_global ('NDPT', obj%ndpt)
    call pc_get_global ('DT', obj%dt)
    call pc_get_global ('GATHERED', gather)
	call pc_get_jdata('DEBUG_LEVEL',obj%debug)
    if (obj%ndpt == inil) call pc_error ("NDPT global hasn't been set.")
    if (obj%ndpt < 512) then
     call pc_error("Need at least 512 samples for a sensible migration")
    end if
	if (obj%ndpt > 4000) then
	 call pc_warning('Trace length too long - resample or truncate')
	end if
    if (obj%dt   == fnil) call pc_error ("DT global hasn't been set.")
    if (.NOT. gather) call pc_warning('GATHERED traces are required')
    call pc_get ('PHASE45'  , obj%phase45)
    call pc_get ('MAXANG'   ,obj%maxang )
    call pc_get ('VEL_PC'   ,obj%vel_pc  )
    call pc_get ('CDPINC'	  ,obj%cdpinc )
    call pc_get ('APERTUR'	   ,obj%apertur )
    call pc_get ('VEL_PATHNAME', obj%vel_pathname)

    !!------------------------- verify parameters --------------------------!!
    !!------------------------- verify parameters --------------------------!!
    !!------------------------- verify parameters --------------------------!!
    call pathcheck &
            ('vel_pathname', obj%vel_pathname, show=PATHCHECK_INFO_INPUT)
! add some pc_warning() here
	badinp=0
	if (obj%maxang <1.0 .or. obj%maxang >80.0) then
	 obj%maxang =40.0
	 badinp=1
	end if
	if (obj%vel_pc  <90.0 .or. obj%vel_pc  >110.0) then
	 obj%vel_pc  =100.0
	 badinp=1
	end if
	if (obj%cdpinc <1.0 .or. obj%cdpinc >999.0) then
	 obj%cdpinc  = 25.0
	 badinp=1
	end if
	if (obj%apertur <250.0 .or. obj%apertur >9999.9) then
	 obj%apertur  = 3000.0
	 badinp=1
	end if
	if (badinp > 0) call pc_info('Out of range parameter reset')

    !!------------------------ write parameters ----------------------------!!
    !!------------------------ write parameters ----------------------------!!
    !!------------------------ write parameters ----------------------------!!
    ! set up for MPI
    call pc_put_control ('PARALLEL_SAFE'         , .true.)
    call pc_put_control ('PCPS_SEND_MODE'        , 'PCPS_SEND_FIRST_AVAIL')
    call pc_put_control ('PCPS_RECEIVE_MODE'     , 'PCPS_RECEIVE_PASSTHRU')
    call pc_put_control ('PCPS_BUNCH_MODE'       , 'PCPS_BUNCH_TRACE_GROUPS')
    call pc_put_control ('PCPS_SEND_EOF_MODE'    , 'PCPS_SEND_ALL_EOF')
    call pc_put_control ('PCPS_ALT_SEND_MODE'    , 'PCPS_SEND_ALL')
    call pc_put_control ('PCPS_ALT_RECEIVE_MODE' , 'PCPS_RECEIVE_ALL_EOF')
    !
    call pc_put_options_field ('PHASE45', rev_options, rev_noptions)
    call pc_put ('PHASE45'  , obj%phase45)
    call pc_put ('MAXANG',  obj%maxang )
    call pc_put ('VEL_PC',  obj%vel_pc  )
    call pc_put ('CDPINC'	,obj%cdpinc )
    call pc_put ('APERTUR'	,obj%apertur )
    call pc_put ('VEL_PATHNAME', obj%vel_pathname)
! is this correct place for nmo parameters?
    call pc_clear()
    call pc_put( 'PATHNAME', obj%vel_pathname )
    call pc_put( 'OPT_NMO' , 'STK_VEL' )
! want RMS to trace, other parameters are not needed
    if (associated(obj%nmov)) then
           call nmo_update (obj%nmov)
        else
           call nmo_create (obj%nmov)
     endif
     call pc_restore()

    !!----------------------- prepare for execution ------------------------!!
    !!----------------------- prepare for execution ------------------------!!
    !!----------------------- prepare for execution ------------------------!!
      if (associated(obj%fftrc_obj))  call fft_delete (obj%fftrc_obj)
      if (associated(obj%fftcr_obj))  call fft_delete (obj%fftcr_obj)

    if (pc_do_not_process_traces()) return
    !
    obj%skip_wrapup = .false.     ! to run wrapup code after processing.
!.... initialize fft tables.
      obj%nfft=fft_nfctr(obj%ndpt+1)
      ier =  fft_create (obj%fftrc_obj,  1, obj%nfft, 'rtoc')
      if (ier/=0) call pc_error ('Error creating rtoc FFT object')
      ier =  fft_create (obj%fftcr_obj, -1, obj%nfft, 'ctor')
      if (ier/=0) call pc_error ('Error creating ctor FFT object')
!if verbose for this:
    obj%minp=obj%dt/obj%cdpinc
    if (obj%debug > 0) then
     write(obj%plun,*) 'minimum slowness p=',obj%minp
     call fft_print(obj%fftrc_obj,6)
     call fft_print(obj%fftcr_obj,6)
    end if
    if (obj%nfft > 4096) then
	  write(obj%plun,*) "KPSTM2D -- Padded FFT -- is big, please"
	  write(obj%plun,*) "Consider trace resample or truncation"
    endif
    obj%degtap=0.1*obj%maxang+2.0
	obj%freqny= 0.5/obj%dt
!	obj%nclp=MAX(8,NINT(obj%maxang*0.266667))
!	if (obj%nclp > 15) obj%nclp=15
!	lowest=1.5*obj%freqny/obj%nclp
	ifft=obj%nfft/2+1
	lowest=obj%freqny*(ifft/20+6)/ifft
    mt=10*obj%ndpt/1000			! 12 if vel 4-bytes
    ma=(12*obj%nfft+32*(obj%ndpt+2))/1000
    if (obj%debug > 0) then
     write(obj%plun,*) 'KPSTM2D arrays allocation kB: ',ma
     write(obj%plun,*) 'plus kB per trace: ',mt
	 write(obj%plun,*) 'lowest filter Hz = ',lowest
	 write(obj%plun,*) 'angle taper degrees:',obj%degtap
    end if

    if (pc_do_not_process_traces()) return   ! in case of allocation errors.??
    ! ------------------------------------------
    if (obj%vel_pathname /= PATHCHECK_EMPTY) then
	 us=cio_fopen(obj%vel_pathname,'r')
	 if (us == 0) then
	  write(obj%plun,*) 'KPSTM2D  error opening vel file'
	  call pc_warning("failure opening velocity file")
	 end if ! us
	 stat=cio_fclose(us)
    else
	 write(obj%plun,*) 'KPSTM2D  No vel file specified!'
    end if  ! end kpstm2d_pathname /=NONE
! if zero records, probably error message as well

    obj%trace_count             = 0

    !!------------------------- finish update ------------------------------!!
    !!------------------------- finish update ------------------------------!!
    !!------------------------- finish update ------------------------------!!

  end subroutine kpstm2d_update

  !!------------------------------- traps ----------------------------------!!
  !!------------------------------- traps ----------------------------------!!
  !!------------------------------- traps ----------------------------------!!


  !!--------------------------- main execution -----------------------------!!
  !!--------------------------- main execution -----------------------------!!
  !!--------------------------- main execution -----------------------------!!

  subroutine kpstm2d (obj,ntr,hd,tr)
    !
    ! - Arguments
    !
    type (kpstm2d_struct), intent (inout) :: obj
    integer,               intent (inout) :: ntr
    double precision,      intent (inout) :: hd (:,:)
    real             ,     intent (inout) :: tr (:,:)
    !
    ! - Local variables
    real	::	vrms(obj%ndpt,1)	! for nmo call
    !
	integer :: aaplo, aaphi
	INTEGER :: i,iaaf,imp,it,ix,jip,k
	integer :: ilx,iux,it1,ic
	integer :: ih,il,iz,ntn,itop
	INTEGER :: nfft,nf	  ! number of FFT components
	INTEGER :: ncmp	      ! number of cdps

	REAL :: anga		! aperture angle
	REAL :: geomsf		! geometrical spreading factor
	REAL :: h	    	! (half) offset
	real :: maxo, mino,maxv,minv
	REAL :: mpx,ipx		! mid-point and image-point location
	REAL :: obliqf		! obliquity correction factor
	REAL :: p	    	! slowness of the Kirchhoff operator
	REAL :: t	    	! time (raypath) 
	REAL :: t0	    	! zero-offset traveltime
	REAL :: tmax0		! maximum time the end
	REAL :: angtaper	! aperture-angle taper
	REAL :: v,vsq,vmul		! velocity 
	REAL :: x	    	! aperture foot/metre
	REAL :: xmaxad		! maximum aperture distance
	REAL :: plustap,tanmax
	REAL :: tts,ttr,t1sq
	REAL :: fraclo,frachi,scala
    REAL :: deg2rad,rad2deg, y,lav, slope
    COMPLEX :: rot45
	REAL :: firstt, amplo, amphi,xt1,tdt
    REAL,DIMENSION(:,:),ALLOCATABLE :: lowpass
    REAL,DIMENSION(:,:),ALLOCATABLE :: migk ! Kirchhoff sum
    REAL,DIMENSION(:),ALLOCATABLE :: trcin	! trace in
    REAL,DIMENSION(:),ALLOCATABLE :: trcout	! filtered trace
    INTEGER(KIND=2),DIMENSION(:,:),ALLOCATABLE :: vel
    COMPLEX,DIMENSION(:),ALLOCATABLE :: ct
    COMPLEX,DIMENSION(:),ALLOCATABLE :: cfft
	DOUBLE PRECISION,DIMENSION(64,1) :: heds ! 2 dimensional, even if 1 trace
	REAL,DIMENSION(4) :: x4lo, x4hi
! assorted trigonometric factors
    rot45=cmplx(0.707107,0.707107) ! phase correction
    rad2deg=57.29578
    deg2rad=0.01745329
    if (ntr == NO_MORE_TRACES .or. ntr == FATAL_ERROR) then
      call kpstm2d_wrapup (obj)
      return
    end if
! should test for small ntr as well
	if (ntr > DIMX) then
	 write(obj%plun,*) "KPSTM2D **************************************"
	 write(obj%plun,*) "Gather is too large to process ***************"
	 ntr=FATAL_ERROR
	 return
	endif 
	! Get info from obj 
	vmul=0.01*obj%vel_pc
	tmax0=(obj%ndpt-1)*obj%dt              ! seconds
	! assume regular, no missing/dupes
	ncmp=ntr
	! Set up FFT parameters 
	nf = obj%nfft/2 + 1
    scala = 1.0/(obj%nfft)
	! Allocate space, needs > 2.5x input gather size!
    ALLOCATE (lowpass(obj%ndpt+2,obj%nclp))
    ALLOCATE (migk(obj%ndpt,ntr))
    ALLOCATE (vel(obj%ndpt,ncmp)) ! ntr=ncmp?
    ALLOCATE (trcin(obj%nfft))
    ALLOCATE (trcout(obj%nfft))
    ALLOCATE (ct(nf))
    ALLOCATE (cfft(nf))

! Wipe all arrays is this needed?
    migk(1:,1:)=0.0
    lowpass(1:,1:)=0.0
	maxo=0.0
	mino=9.9
! fetch the velocity for each CDP
	maxv=0.0
	minv=0.0
	plustap=obj%maxang+obj%degtap
	tanmax =tan(plustap*deg2rad)
!
	do imp=1,ntr
	 heds(1:64,1)=hd(1:64,imp)
	 ntn=1
	 call nmo(obj%nmov,ntn,heds,vrms)
	 minv=minv+vrms(1,1)
	 maxv=maxv+vrms(obj%ndpt,1)
	 vel(1:obj%ndpt,imp)=NINT(vrms(1:obj%ndpt,1)*vmul)
	end do
	minv=minv/ntr	! only for debugging
	maxv=maxv/ntr
! may need to smooth velocity?
! begin the number crunching  
! outer Loop over input trace
    do imp=1,ntr
! Do not start at time zero: may get nonsense=NaN/Inf
		itop=MAX(2,nint(hd(2,imp)/2))
!	half offset
		h=real(abs(hd(6,imp)))*0.5
		mpx=imp*obj%cdpinc               ! distance from start
		if (imp==ntr .AND. obj%debug > 0) then
		 obj%gc=obj%gc+1
		 write(obj%plun,*) "migrated gather ",obj%gc
		end if
		! Compute low-pass filtered siblings  
		! of the trace to be used for antialiasing 
		do it=1,obj%ndpt
			trcin(it)= tr(it,imp)
		end do
! ensure zero-padded
		trcin(obj%ndpt+1:obj%nfft)=0.0
        call fft_rc_transform(obj%fftrc_obj, trcin, cfft)
        if (obj%phase45=='YES') then
		 do it=1,nf
			cfft(it) = cfft(it)*rot45
		 end do
		end if
		lowpass(1:obj%ndpt+1,obj%nclp)=trcin(1:obj%ndpt+1)	! last is unfiltered
! This fudges the low-pass frequencies slightly (e.g. 1.5 Hz for 4 second records).
! The lowest low-pass may choke most energy, so beyond 60 degrees little passes.
! In practice this relaxed filtering seems to do no harm.
! if unalias=no, omit this loop?
		do iaaf=1,obj%nclp-1
!		  ic=int((2*iaaf+1)*nf/obj%nclp/2)+1
		  ic=iaaf*nf/obj%nclp+6
		  il=ic-iaaf-1					! filter edge
		  ih=ic+iaaf+1				! filter edge
		  iz=ih+1
		  slope=3.14159/real(ih-il+2)
		  ct(1:ih+1)=cfft(1:ih+1)
! Hann taper
		  do k=il,ih
		   ct(k)=ct(k)*(1.0-cos((iz-k)*slope))*0.5
		  end do
	
		  ct(iz:nf)=cmplx(0.0,0.0)
          call fft_cr_transform(obj%fftcr_obj, ct, trcout, scala) !
			do it=1,obj%ndpt+1
				lowpass(it,iaaf)= trcout(it)  !lowpass(t,f)
			end do
		end do  ! iaaf
! lowpass is current trace, smeared into surrounding output traces
		! Loop over vertical 2WT 
		 do it=itop,obj%ndpt		! skip half of top mute
! imp starts at 1
			t0=(it-1)*obj%dt	! ignoring TSTRT
            t1sq=t0*t0*0.25
			v=REAL(vel(it,imp))
            vsq=v*v
! tan => rapidly stretching aperture as angle increased
			xmaxad=MIN(tanmax*v*t0,obj%apertur)
			ilx=MAX(1,imp - nint(xmaxad/obj%cdpinc))  !
			iux=MIN(ntr,imp + nint(xmaxad/obj%cdpinc)) !aperture limits
! scan across output points to the west of the midpoint
		  do jip=imp,ilx,-1
			ipx=jip*obj%cdpinc	!!!
			x=ipx-mpx
			tts=sqrt(t1sq + ((x+h)/v)**2 )
			ttr=sqrt(t1sq + ((h-x)/v)**2 )
			t= tts + ttr
			if (t>=tmax0)  exit
! geometric spreading weighting factor based on horizontal layering,
! therefore approximate in real world. But error will be small.
! obliqf is almost 1 at centre, decreases to approximately 0.2 for 70 degrees
! or 0.7 for 40 degrees maximum angle. An alternative aperture limit is to use
! obliqf as a cutoff (for next version?)
	  		obliqf=sqrt(0.5*(1.0+(t1sq/(tts*ttr))-sqrt((tts*tts-t1sq)*(ttr*ttr-t1sq))/(tts*ttr)))
	  		!          1    2    3     4              56              7               8
			geomsf=sqrt(1.0/(t*v))  !
! anga increases rapidly away from centre for shallow data
! Asymptotically approaches 90 degrees, although limits be may be 85 degrees at
! 0.5 sec, or only 70 degrees at 1 sec depending on horizontal aperture
	  		anga=rad2deg*(acos(t0/t))   !57.2958
	  		if (anga<=obj%maxang) angtaper=1.0
			if (anga>obj%maxang) angtaper=(plustap-anga)/obj%degtap	! linear
! calculate p (migration operator slowness) to select
! the nearest low-pass filtered trace for antialiasing
			p=abs((x+h)/(vsq*tts) + (x-h)/(vsq*ttr))
! if p<Pmin then aaplo will be maximal
! p increases away from centre, so aaplo decreases. For shallow times, aaplo
! may jump by 2 or 3 on traces close to centre, but that will correspond to
! frequencies > 2/3 of Nyquist
			if (p>obj%minp) then
                 aaplo=max(int(obj%nclp*obj%minp/p),1)
			else
                 aaplo=obj%nclp !  no aliasing
			end if
				frachi=amod(obj%nclp*obj%minp,p)
				fraclo=1.0-frachi
				aaphi=aaplo+1 !
! t=travel time; it1=index of sample immediately lesser
                tdt=t/obj%dt
                it1=int(tdt)		! index
                xt1=tdt-it1			! fraction of a sample
! swing energy including fraction of sample
				if (angtaper > 0.0) then
	  		    if (obliqf<mino) mino=obliqf
	  		    if (obliqf>maxo) maxo=obliqf
				if (aaplo>=obj%nclp) then	! horizomble
					do k=1,4
						x4lo(k)=lowpass(k+it1,obj%nclp) ! use unfiltered
					end do
                    if (it1 >= obj%ndpt) then
                     amplo=lowpass(obj%ndpt,obj%nclp)
                    else ! cubic spline interpolation
                     call cubint(x4lo,xt1,amplo)
                    end if ! it1
					migk(it,jip) =migk(it,jip)+geomsf*obliqf*angtaper*amplo
				 else ! if (aaplo<obj%nclp)
					do k=1,4
						x4lo(k)=lowpass(k+it1,aaplo) ! use nearest lowpass
						x4hi(k)=lowpass(k+it1,aaphi) !
					end do
                    if (it1 >= obj%ndpt) then
                     amplo=lowpass(obj%ndpt,aaplo)
                     amphi=lowpass(obj%ndpt,aaphi)
                    else ! cubic spline, since mostly lowered frequencies
                     call cubint(x4lo,xt1,amplo)
                     call cubint(x4hi,xt1,amphi)
                    end if
					migk(it,jip)=migk(it,jip)+ geomsf*obliqf*angtaper*(fraclo*amplo + frachi*amphi)
				end if ! aaplo
				end if ! angtaper
			end do ! loop jip
! span across output locations to the east of the midpoint
! note imp+1
			do jip=imp+1,iux
				ipx=jip*obj%cdpinc
				x=ipx-mpx
				tts=sqrt( t1sq + ((x+h)/v)**2 )
				ttr=sqrt( t1sq + ((h-x)/v)**2 )
				t= tts + ttr
				if (t>=tmax0)  exit ! out of jip loop
				obliqf=sqrt(0.5*(1 + (t1sq/(tts*ttr))-sqrt((tts*tts-t1sq)*(ttr*ttr- t1sq))/(tts*ttr)))
				geomsf=sqrt(1.0/(t*v))			
				anga=rad2deg*abs(acos(t0/t))
! saw no advantage with cosine taper over linear ramp
				if (anga<=obj%maxang) angtaper=1.0
				if (anga>obj%maxang) angtaper=(plustap-anga)/obj%degtap	! linear
	  		! calculate p (migration operator slowness) to select
			! the nearest low-pass filtered trace for antialiasing
				p=abs((x+h)/(vsq*tts) + (x-h)/(vsq*ttr))
				if (p>obj%minp) then
                 aaplo=max(int(obj%nclp*obj%minp/p),1)
                else
                 aaplo=obj%nclp !  no aliasing
                end if
				frachi=amod(obj%nclp*obj%minp,p) ! remainder
				fraclo=1.0-frachi
				aaphi=aaplo+1 !
                tdt=t/obj%dt		! use this
                it1=int(tdt)		! index
                xt1=tdt-it1			! fraction of a sample
				! swing sample from CMP to CIP
				if (angtaper > 0.0) then
	  		    if (obliqf<mino) mino=obliqf
	  		    if (obliqf>maxo) maxo=obliqf
				if (aaplo>=obj%nclp) then
					do k=1,4
						x4lo(k)=lowpass(k+it1,obj%nclp) ! use unfiltered
					end do
                    if (it1 >= obj%ndpt) then
                     amplo=lowpass(obj%ndpt,obj%nclp)
                    else ! splines
                     call cubint(x4lo,xt1,amplo)
                    end if
					migk(it,jip) =migk(it,jip)+geomsf*obliqf*angtaper*amplo
				 else !(aaplo<obj%nclp)
					do k=1,4
						x4lo(k)=lowpass(k+it1,aaplo) !
						x4hi(k)=lowpass(k+it1,aaphi) ! s
					end do
                    call cubint(x4lo,xt1,amplo)
                    call cubint(x4hi,xt1,amphi)
                    !      n  dx   fxin   yin     extrapol         output
                    if (it1 >= obj%ndpt) then
                     amplo=lowpass(obj%ndpt,aaplo)
                     amphi=lowpass(obj%ndpt,aaphi)
                    else ! crude linear interp, use cubic later
                     call cubint(x4lo,xt1,amplo)
                     call cubint(x4hi,xt1,amphi)
                    end if
					migk(it,jip) =migk(it,jip)+ geomsf*obliqf*angtaper*(fraclo*amplo + frachi*amphi)
				end if ! aaplo
				end if ! angtaper
			end do ! loop jip
		end do ! time loop
	end do ! trace loop
	if (obj%debug > 0) write(obj%plun,*) '[info] min max obliqf: ',mino,maxo
	! Output a migrated gather and set LAV
	do ix=1,ntr
		lav=0.0
		do it=1,obj%ndpt-1
			y = migk(it,ix)
			if (abs(y) > lav) lav=abs(y)
			tr(it+1,ix) = y				! CHECK THIS it+1
		end do
		hd(25,ix)=lav
!  update mute headers?
	end do
!
    DEALLOCATE (lowpass)
    DEALLOCATE (migk)
    DEALLOCATE (vel)
    DEALLOCATE (trcin)	! could perhaps make these last 4 fixed arrays
    DEALLOCATE (trcout)	! to reduce alloc/dealloc?
    DEALLOCATE (ct)
    DEALLOCATE (cfft)

    if (ntr > 0) obj%trace_count = obj%trace_count + ntr
    !
    !
  end subroutine kpstm2d

  !!--------------------------- kpstm2d_wrapup -----------------------------!!
  !!--------------------------- kpstm2d_wrapup -----------------------------!!
  !!--------------------------- kpstm2d_wrapup -----------------------------!!

  subroutine kpstm2d_wrapup (obj)
    !
    ! - Arguments
    !
    type (kpstm2d_struct), intent (inout) :: obj       ! arguments
    !
    !
    ! - Begin kpstm2d_wrapup
    !
    if (obj%skip_wrapup) return
    obj%skip_wrapup = .true.
    !
    ! - List statistics
    !
    if (obj%debug > 0) then
     write(obj%plun,*) '-------wrapup phase---'
     write(obj%plun,*) 'KPSTM2D:  Total traces= ', obj%trace_count
    end if
    !
  end subroutine kpstm2d_wrapup

  !!--------------------------- end of module ------------------------------!!
  !!--------------------------- end of module ------------------------------!!
  !!--------------------------- end of module ------------------------------!!

subroutine cubint(y,mu,yout)
! given an array of 4 points, interpolate between 2nd and 3rd by cubic spline
	  real,intent(in)	:: y(1:4)
	  real,intent(in)	:: mu		! fraction of a sample
	  real,intent(out)	:: yout		! interpolated value
	  real				:: a1,a2,a3
	  integer			:: i
	  i=2
	  a3=0.5*(y(i+2)-y(i-1))+1.5*(y(i)-y(i+1))
	  a2=y(i-1)-2.5*y(i)+2.0*y(i+1)-0.5*y(i+2)
	  a1=0.5*(y(i+1)-y(i-1))
	  yout=((a3*mu+a2)*mu+a1)*mu+y(i)
	  return
	  end subroutine cubint

end module kpstm2d_module

!!--------------------------------- end ------------------------------------!!
!!--------------------------------- end ------------------------------------!!
!!--------------------------------- end ------------------------------------!!
